from unittest import TestCase
import numpy as np
import pandas as pd
import itertools
import pytest

from mavedb.lib.validation.exceptions import ValidationError
from mavedb.lib.validation.constants.general import (
    hgvs_nt_column,
    hgvs_pro_column,
    hgvs_splice_column,
    required_score_column,
)

from mavedb.lib.validation.dataframe import (
    infer_column_type,
    sort_dataframe_columns,
    standardize_dataframe,
    validate_no_null_rows,
    validate_column_names,
    validate_hgvs_column,
    validate_hgvs_prefix_combinations,
    validate_data_column,
    validate_variant_columns_match,
)


class TestInferColumnType(TestCase):
    def test_floats(self):
        test_data = pd.Series([12.0, 1.0, -0.012, 5.75])
        self.assertEqual(infer_column_type(test_data), "numeric")

    def test_ints(self):
        test_data = pd.Series([12, 1, 0, -5])
        self.assertEqual(infer_column_type(test_data), "numeric")

    def test_floats_with_na(self):
        test_data = pd.Series([12.0, 1.0, None, -0.012, 5.75])
        self.assertEqual(infer_column_type(test_data), "numeric")

    def test_ints_with_na(self):
        test_data = pd.Series([12, 1, None, 0, -5])
        self.assertEqual(infer_column_type(test_data), "numeric")

    def test_convertable_strings(self):
        test_data = pd.Series(["12.5", 1.25, "0", "-5"])
        self.assertEqual(infer_column_type(test_data), "numeric")

    def test_strings(self):
        test_data = pd.Series(["hello", "test", "suite", "123abc"])
        self.assertEqual(infer_column_type(test_data), "string")

    def test_strings_with_na(self):
        test_data = pd.Series(["hello", "test", None, "suite", "123abc"])
        self.assertEqual(infer_column_type(test_data), "string")

    def test_mixed(self):
        test_data = pd.Series(["hello", 12.123, -75, "123abc"])
        self.assertEqual(infer_column_type(test_data), "mixed")

    def test_mixed_with_na(self):
        test_data = pd.Series(["hello", None, 12.123, -75, "123abc"])
        self.assertEqual(infer_column_type(test_data), "mixed")

    def test_all_na(self):
        test_data = pd.Series([None] * 5)
        self.assertEqual(infer_column_type(test_data), "empty")


class DfTestCase(TestCase):
    def setUp(self):
        self.dataframe = pd.DataFrame(
            {
                hgvs_nt_column: ["g.1A>G", "g.1A>T"],
                hgvs_splice_column: ["c.1A>G", "c.1A>T"],
                hgvs_pro_column: ["p.Met1Val", "p.Met1Leu"],
                required_score_column: [1.0, 2.0],
                "extra": [12.0, 3.0],
                "count1": [3.0, 5.0],
                "count2": [9, 10],
                "extra2": ["pathogenic", "benign"],
            }
        )

        self.target_seq = "ATG"
        self.target_seq_type = "dna"


class TestValidateDataColumn(DfTestCase):
    def test_valid(self):
        validate_data_column(self.dataframe[required_score_column])

    def test_null_column(self):
        self.dataframe[required_score_column] = None
        with self.assertRaises(ValidationError):
            validate_data_column(self.dataframe[required_score_column])

    def test_missing_data(self):
        self.dataframe.loc[0, "extra"] = None
        validate_data_column(self.dataframe["extra"])

    def test_force_numeric(self):
        with self.assertRaises(ValidationError):
            validate_data_column(self.dataframe["extra2"], force_numeric=True)

    def test_mixed_types_invalid(self):
        with self.assertRaises(ValidationError):
            validate_data_column(pd.Series(["test", 1.0]))


class TestNullRows(DfTestCase):
    def test_valid(self):
        validate_no_null_rows(self.dataframe)

    def test_null_row(self):
        self.dataframe.loc[1] = None
        with self.assertRaises(ValidationError):
            validate_no_null_rows(self.dataframe)

    def test_only_hgvs_row(self):
        self.dataframe.loc[1, [required_score_column, "extra", "count1", "count2"]] = None
        validate_no_null_rows(self.dataframe)


class TestColumnNames(DfTestCase):
    def test_only_two_kinds_of_dataframe(self):
        with self.assertRaises(ValueError):
            validate_column_names(self.dataframe, kind="score2")

    def test_require_dataframe_kind(self):
        with self.assertRaises(TypeError):
            validate_column_names(self.dataframe)

    def test_score_df_has_score_column(self):
        validate_column_names(self.dataframe, kind="scores")
        with self.assertRaises(ValidationError):
            validate_column_names(self.dataframe.drop([required_score_column], axis=1), kind="scores")

    def test_count_df_lacks_score_column(self):
        validate_column_names(self.dataframe.drop([required_score_column], axis=1), kind="counts")
        with self.assertRaises(ValidationError):
            validate_column_names(self.dataframe, kind="counts")

    def test_df_with_only_scores(self):
        validate_column_names(self.dataframe[[hgvs_pro_column, required_score_column]], kind="scores")

    def test_count_df_must_have_data(self):
        with self.assertRaises(ValidationError):
            validate_column_names(self.dataframe[[hgvs_nt_column, hgvs_pro_column]], kind="counts")

    def test_just_hgvs_nt(self):
        validate_column_names(self.dataframe.drop([hgvs_pro_column, hgvs_splice_column], axis=1), kind="scores")
        validate_column_names(
            self.dataframe.drop([hgvs_pro_column, hgvs_splice_column, required_score_column], axis=1), kind="counts"
        )

    def test_just_hgvs_pro(self):
        validate_column_names(self.dataframe.drop([hgvs_nt_column, hgvs_splice_column], axis=1), kind="scores")
        validate_column_names(
            self.dataframe.drop([hgvs_nt_column, hgvs_splice_column, required_score_column], axis=1), kind="counts"
        )

    def test_just_hgvs_pro_and_nt(self):
        validate_column_names(self.dataframe.drop([hgvs_splice_column], axis=1), kind="scores")
        validate_column_names(self.dataframe.drop([hgvs_splice_column, required_score_column], axis=1), kind="counts")

    def test_hgvs_splice_must_have_pro_and_nt(self):
        with self.assertRaises(ValidationError):
            validate_column_names(self.dataframe.drop([hgvs_nt_column, hgvs_pro_column], axis=1), kind="scores")
        with self.assertRaises(ValidationError):
            validate_column_names(self.dataframe.drop([hgvs_nt_column], axis=1), kind="scores")
        with self.assertRaises(ValidationError):
            validate_column_names(self.dataframe.drop([hgvs_pro_column], axis=1), kind="scores")
        with self.assertRaises(ValidationError):
            validate_column_names(
                self.dataframe.drop([hgvs_nt_column, hgvs_pro_column, required_score_column], axis=1), kind="counts"
            )
        with self.assertRaises(ValidationError):
            validate_column_names(self.dataframe.drop([hgvs_nt_column, required_score_column], axis=1), kind="counts")
        with self.assertRaises(ValidationError):
            validate_column_names(self.dataframe.drop([hgvs_pro_column, required_score_column], axis=1), kind="counts")

    def test_no_hgvs_column(self):
        with pytest.raises(ValidationError) as exc_info:
            validate_column_names(
                self.dataframe.drop([hgvs_nt_column, hgvs_pro_column, hgvs_splice_column], axis=1), kind="scores"
            )
        assert "dataframe does not define any variant columns" in str(exc_info.value)
        with pytest.raises(ValidationError) as exc_info:
            validate_column_names(
                self.dataframe.drop(
                    [hgvs_nt_column, hgvs_pro_column, hgvs_splice_column, required_score_column], axis=1
                ),
                kind="counts",
            )
        assert "dataframe does not define any variant columns" in str(exc_info.value)

    def test_validation_ignores_column_ordering(self):
        validate_column_names(
            self.dataframe[[hgvs_nt_column, required_score_column, hgvs_pro_column, hgvs_splice_column]], kind="scores"
        )
        validate_column_names(self.dataframe[[required_score_column, hgvs_nt_column, hgvs_pro_column]], kind="scores")
        validate_column_names(self.dataframe[[hgvs_pro_column, required_score_column, hgvs_nt_column]], kind="scores")

        validate_column_names(
            self.dataframe[[hgvs_nt_column, "count1", hgvs_pro_column, hgvs_splice_column, "count2"]], kind="counts"
        )
        validate_column_names(self.dataframe[["count1", "count2", hgvs_nt_column, hgvs_pro_column]], kind="counts")
        validate_column_names(self.dataframe[[hgvs_pro_column, "count1", "count2", hgvs_nt_column]], kind="counts")

    def test_validation_is_case_insensitive(self):
        validate_column_names(self.dataframe.rename(columns={hgvs_nt_column: hgvs_nt_column.upper()}), kind="scores")
        validate_column_names(
            self.dataframe.rename(columns={required_score_column: required_score_column.title()}), kind="scores"
        )

    def test_duplicate_hgvs_column_names(self):
        with self.assertRaises(ValidationError):
            validate_column_names(self.dataframe.rename(columns={hgvs_pro_column: hgvs_nt_column}), kind="scores")
        with self.assertRaises(ValidationError):
            validate_column_names(
                self.dataframe.drop([required_score_column], axis=1).rename(columns={hgvs_pro_column: hgvs_nt_column}),
                kind="counts",
            )

    def test_duplicate_score_column_names(self):
        with self.assertRaises(ValidationError):
            validate_column_names(self.dataframe.rename(columns={"extra": required_score_column}), kind="scores")

    def test_duplicate_data_column_names(self):
        with self.assertRaises(ValidationError):
            validate_column_names(self.dataframe.rename(columns={"count2": "count1"}), kind="scores")
        with self.assertRaises(ValidationError):
            validate_column_names(
                self.dataframe.drop([required_score_column], axis=1).rename(columns={"count2": "count1"}), kind="counts"
            )

    def test_invalid_column_names(self):
        invalid_values = [None, np.nan, "", " "]
        for value in invalid_values:
            with self.subTest(value=value):
                with self.assertRaises(ValidationError):
                    validate_column_names(self.dataframe.rename(columns={hgvs_splice_column: value}), kind="scores")
                with self.assertRaises(ValidationError):
                    validate_column_names(
                        self.dataframe.drop([required_score_column], axis=1).rename(
                            columns={hgvs_splice_column: value}
                        ),
                        kind="counts",
                    )

    def test_ignore_column_ordering(self):
        validate_column_names(
            self.dataframe[[hgvs_splice_column, "extra", "count1", hgvs_pro_column, "score", hgvs_nt_column, "count2"]],
            kind="scores",
        )
        validate_column_names(
            self.dataframe[[hgvs_splice_column, "extra", "count1", hgvs_pro_column, hgvs_nt_column, "count2"]],
            kind="counts",
        )


class TestSortDataframeColumns(DfTestCase):
    def test_preserve_sorted(self):
        sorted_df = sort_dataframe_columns(self.dataframe)
        pd.testing.assert_frame_equal(self.dataframe, sorted_df)

    def test_sort_dataframe(self):
        sorted_df = sort_dataframe_columns(
            self.dataframe[
                [
                    hgvs_splice_column,
                    "extra",
                    "count1",
                    hgvs_pro_column,
                    required_score_column,
                    hgvs_nt_column,
                    "count2",
                    "extra2",
                ]
            ]
        )
        pd.testing.assert_frame_equal(self.dataframe, sorted_df)

    def test_sort_dataframe_is_case_insensitive(self):
        self.dataframe = self.dataframe.rename(columns={hgvs_nt_column: hgvs_nt_column.upper()})
        sorted_df = sort_dataframe_columns(self.dataframe)
        pd.testing.assert_frame_equal(self.dataframe, sorted_df)

    def test_sort_dataframe_preserves_extras_order(self):
        sorted_df = sort_dataframe_columns(
            self.dataframe[
                [
                    hgvs_splice_column,
                    "count2",
                    hgvs_pro_column,
                    required_score_column,
                    hgvs_nt_column,
                    "count1",
                    "extra2",
                    "extra",
                ]
            ]
        )
        pd.testing.assert_frame_equal(
            self.dataframe[
                [
                    hgvs_nt_column,
                    hgvs_splice_column,
                    hgvs_pro_column,
                    required_score_column,
                    "count2",
                    "count1",
                    "extra2",
                    "extra",
                ]
            ],
            sorted_df,
        )


class TestStandardizeDataframe(DfTestCase):
    def test_preserve_standardized(self):
        standardized_df = standardize_dataframe(self.dataframe)
        pd.testing.assert_frame_equal(self.dataframe, standardized_df)

    def test_standardize_changes_case(self):
        standardized_df = standardize_dataframe(self.dataframe.rename(columns={hgvs_nt_column: hgvs_nt_column.upper()}))
        pd.testing.assert_frame_equal(self.dataframe, standardized_df)
        standardized_df = standardize_dataframe(
            self.dataframe.rename(columns={required_score_column: required_score_column.title()})
        )
        pd.testing.assert_frame_equal(self.dataframe, standardized_df)

    def test_standardize_preserves_extras_case(self):
        standardized_df = standardize_dataframe(self.dataframe.rename(columns={"extra": "extra".upper()}))
        pd.testing.assert_frame_equal(self.dataframe.rename(columns={"extra": "extra".upper()}), standardized_df)

    def test_standardize_sorts_columns(self):
        standardized_df = standardize_dataframe(
            self.dataframe[
                [
                    hgvs_splice_column,
                    "count2",
                    hgvs_pro_column,
                    required_score_column,
                    hgvs_nt_column,
                    "count1",
                    "extra",
                ]
            ]
        )
        pd.testing.assert_frame_equal(
            self.dataframe[
                [
                    hgvs_nt_column,
                    hgvs_splice_column,
                    hgvs_pro_column,
                    required_score_column,
                    "count2",
                    "count1",
                    "extra",
                ]
            ],
            standardized_df,
        )


class TestValidateHgvsColumn(DfTestCase):
    def setUp(self):
        super().setUp()

        self.valid_hgvs_columns = [
            pd.Series(["g.1A>G", "g.1A>T"], name=hgvs_nt_column),
            pd.Series(["m.1A>G", "m.1A>T"], name=hgvs_nt_column),
            pd.Series(["c.1A>G", "c.1A>T"], name=hgvs_nt_column),
            pd.Series(["n.1A>G", "n.1A>T"], name=hgvs_nt_column),
            pd.Series(["c.1A>G", "c.1A>T"], name=hgvs_splice_column),
            pd.Series(["p.Met1Val", "p.Met1Leu"], name=hgvs_pro_column),
        ]

        self.valid_hgvs_columns_invalid_for_index = [
            # missing data
            pd.Series(["c.1A>G", None], name=hgvs_nt_column),
            pd.Series([None, "p.Met1Val"], name=hgvs_pro_column),
            pd.Series([None, None], name=hgvs_nt_column),
            pd.Series([None, None], name=hgvs_pro_column),
            # duplicate rows
            pd.Series(["c.1A>G", "c.1A>G"], name=hgvs_nt_column),
            pd.Series(["p.Met1Val", "p.Met1Val"], name=hgvs_pro_column),
        ]

        self.invalid_hgvs_columns_by_name = [
            pd.Series(["g.1A>G", "g.1A>T"], name=hgvs_splice_column),
            pd.Series(["g.1A>G", "g.1A>T"], name=hgvs_pro_column),
            pd.Series(["c.1A>G", "c.1A>T"], name=hgvs_pro_column),
            pd.Series(["n.1A>G", "n.1A>T"], name=hgvs_pro_column),
            pd.Series(["p.Met1Val", "p.Met1Leu"], name=hgvs_nt_column),
        ]

        self.invalid_hgvs_columns_by_contents = [
            pd.Series(["r.1a>g", "r.1a>u"], name=hgvs_splice_column),  # rna not allowed
            pd.Series(["r.1a>g", "r.1a>u"], name=hgvs_nt_column),  # rna not allowed
            pd.Series(["c.1A>G", "c.5A>T"], name=hgvs_nt_column),  # out of bounds for target
            pd.Series(["c.1A>G", "_wt"], name=hgvs_nt_column),  # old special variant
            pd.Series(["p.Met1Leu", "_sy"], name=hgvs_pro_column),  # old special variant
            pd.Series(["n.1A>G", "c.1A>T"], name=hgvs_nt_column),  # mixed prefix
            pd.Series(["c.1A>G", "p.Met1Leu"], name=hgvs_pro_column),  # mixed types/prefix
            pd.Series(["c.1A>G", 2.5], name=hgvs_nt_column),  # contains numeric
            pd.Series([1.0, 2.5], name=hgvs_nt_column),  # contains numeric
            pd.Series([1.0, 2.5], name=hgvs_splice_column),  # contains numeric
            pd.Series([1.0, 2.5], name=hgvs_pro_column),  # contains numeric
        ]

    def test_valid_columns(self):
        for column in self.valid_hgvs_columns:
            with self.subTest(column=column):
                validate_hgvs_column(
                    column, is_index=False, target_seq=self.target_seq, target_seq_type=self.target_seq_type
                )
        for column in self.valid_hgvs_columns_invalid_for_index:
            with self.subTest(column=column):
                validate_hgvs_column(
                    column, is_index=False, target_seq=self.target_seq, target_seq_type=self.target_seq_type
                )

    def test_index_columns(self):
        for column in self.valid_hgvs_columns:
            with self.subTest(column=column):
                validate_hgvs_column(
                    column, is_index=True, target_seq=self.target_seq, target_seq_type=self.target_seq_type
                )
        for column in self.valid_hgvs_columns_invalid_for_index:
            with self.subTest(column=column):
                with self.assertRaises(ValidationError):
                    validate_hgvs_column(
                        column, is_index=True, target_seq=self.target_seq, target_seq_type=self.target_seq_type
                    )

    def test_invalid_column_values(self):
        for column in self.invalid_hgvs_columns_by_contents:
            with self.subTest(column=column):
                with self.assertRaises(ValidationError):
                    validate_hgvs_column(
                        column, is_index=False, target_seq=self.target_seq, target_seq_type=self.target_seq_type
                    )
        for column in self.invalid_hgvs_columns_by_contents:
            with self.subTest(column=column):
                with self.assertRaises(ValidationError):
                    validate_hgvs_column(
                        column, is_index=True, target_seq=self.target_seq, target_seq_type=self.target_seq_type
                    )

    def test_valid_column_values_wrong_column_name(self):
        for column in self.invalid_hgvs_columns_by_name:
            with self.subTest(column=column):
                with self.assertRaises(ValidationError):
                    validate_hgvs_column(
                        column, is_index=False, target_seq=self.target_seq, target_seq_type=self.target_seq_type
                    )
        for column in self.invalid_hgvs_columns_by_name:
            with self.subTest(column=column):
                with self.assertRaises(ValidationError):
                    validate_hgvs_column(
                        column, is_index=True, target_seq=self.target_seq, target_seq_type=self.target_seq_type
                    )


class TestValidateHgvsPrefixCombinations(TestCase):
    def setUp(self):
        self.valid_combinations = [
            ("g", "c", "p"),
            ("m", "c", "p"),
            ("o", "c", "p"),
            ("g", "n", None),
            ("m", "n", None),
            ("o", "n", None),
            ("n", None, None),
            ("c", None, "p"),
            (None, None, "p"),
            (None, None, None),  # valid for this validator, but a dataframe with no variants should be caught upstream
        ]
        self.invalid_combinations = [
            t
            for t in itertools.product(list("cngmo") + [None], ("c", "n", None), ("p", None))
            if t not in self.valid_combinations
        ]

    def test_valid_combinations(self):
        for t in self.valid_combinations:
            with self.subTest(t=t):
                validate_hgvs_prefix_combinations(*t)

    def test_invalid_combinations(self):
        for t in self.invalid_combinations:
            with self.subTest(t=t):
                with self.assertRaises(ValidationError):
                    validate_hgvs_prefix_combinations(*t)

    def test_invalid_combinations_value_error(self):
        with self.assertRaises(ValueError):
            validate_hgvs_prefix_combinations("p", None, None)
        with self.assertRaises(ValueError):
            validate_hgvs_prefix_combinations("c", None, "P")
        with self.assertRaises(ValueError):
            validate_hgvs_prefix_combinations("x", "c", "p")


class TestValidateVariantColumnsMatch(DfTestCase):
    def test_same_df(self):
        validate_variant_columns_match(self.dataframe, self.dataframe)

    def test_ignore_order(self):
        validate_variant_columns_match(self.dataframe, self.dataframe.iloc[::-1])

    def test_missing_column(self):
        with self.assertRaises(ValidationError):
            validate_variant_columns_match(self.dataframe, self.dataframe.drop(hgvs_nt_column, axis=1))
        with self.assertRaises(ValidationError):
            validate_variant_columns_match(self.dataframe.drop(hgvs_nt_column, axis=1), self.dataframe)

    def test_missing_variant(self):
        df2 = self.dataframe.copy()
        df2.loc[0, hgvs_pro_column] = None
        with self.assertRaises(ValidationError):
            validate_variant_columns_match(self.dataframe, df2)



from typing import Any, List

from fastapi import APIRouter, Depends, HTTPException
from sqlalchemy.orm import Session

from mavedb import deps
from mavedb.models.reference_genome import ReferenceGenome
from mavedb.view_models import reference_genome

router = APIRouter(
    prefix="/api/v1/referenceGenomes", tags=["reference-genomes"], responses={404: {"description": "Not found"}}
)


@router.get("/", status_code=200, response_model=List[reference_genome.ReferenceGenome], responses={404: {}})
def list_reference_genomes(
    *,
    db: Session = Depends(deps.get_db),
) -> Any:
    """
    List reference genomes.
    """
    items = db.query(ReferenceGenome).order_by(ReferenceGenome.short_name).all()
    return items


@router.get("/organismNames", status_code=200, response_model=List[str], responses={404: {}})
def list_reference_genome_organism_names(
    *,
    db: Session = Depends(deps.get_db),
) -> Any:
    """
    List distinct reference genome organism names, in alphabetical order.
    """

    items = db.query(ReferenceGenome).all()
    organism_names = map(lambda item: item.organism_name, items)
    return sorted(list(set(organism_names)))


@router.get("/{item_id}", status_code=200, response_model=reference_genome.ReferenceGenome, responses={404: {}})
def fetch_reference_genome(
    *,
    item_id: int,
    db: Session = Depends(deps.get_db),
) -> Any:
    """
    Fetch a single reference genome by ID.
    """
    item = db.query(ReferenceGenome).filter(ReferenceGenome.id == item_id).first()
    if not item:
        raise HTTPException(status_code=404, detail=f"ReferenceGenome with ID {item_id} not found")
    return item



# From https://improveandrepeat.com/2021/09/python-friday-87-handling-pre-existing-tables-with-alembic-and-sqlalchemy/
# Based on https://github.com/talkpython/data-driven-web-apps-with-flask

from alembic import op
from sqlalchemy import engine_from_config
from sqlalchemy.engine import reflection


def table_does_not_exist(table, schema=None):
    config = op.get_context().config
    engine = engine_from_config(config.get_section(config.config_ini_section), prefix="sqlalchemy.")
    insp = reflection.Inspector.from_engine(engine)
    return insp.has_table(table, schema) is False


def table_has_column(table, column):
    config = op.get_context().config
    engine = engine_from_config(config.get_section(config.config_ini_section), prefix="sqlalchemy.")
    insp = reflection.Inspector.from_engine(engine)
    has_column = False
    for col in insp.get_columns(table):
        if column not in col["name"]:
            continue
        has_column = True
    return has_column



from mavedb.view_models.wild_type_sequence import WildTypeSequenceCreate

import pytest


@pytest.mark.parametrize(
    "sequence_type, sequence",
    [
        ("dna", "ATGAGTATTCAACATTTCCGTGTC"),
        ("dna", "ATGAGTATtcaatcTTTCCGTGTC"),
        ("protein", "STARTREK"),
        ("Protein", "startrek"),
    ],
)
def test_create_wild_type_sequence(test_empty_db, sequence_type, sequence):
    wildTypeSeq = WildTypeSequenceCreate(sequence_type=sequence_type, sequence=sequence)
    assert wildTypeSeq.sequence_type == sequence_type.lower()
    assert wildTypeSeq.sequence == sequence.upper()


@pytest.mark.parametrize("sequence_type, sequence", [("dnaaa", "ATGAGTATTCAACATTTCCGTGTC"), ("null", "STARTREK")])
def test_create_invalid_sequence_type(test_empty_db, sequence_type, sequence):
    with pytest.raises(ValueError) as exc_info:
        WildTypeSequenceCreate(sequence_type=sequence_type, sequence=sequence)
    assert f"'{sequence_type}' is not a valid sequence type" in str(exc_info.value)


@pytest.mark.parametrize("sequence_type, sequence", [("dna", "ARCG"), ("protein", "AzCG")])
def test_create_invalid_sequence(test_empty_db, sequence_type, sequence):
    with pytest.raises(ValueError) as exc_info:
        WildTypeSequenceCreate(sequence_type=sequence_type, sequence=sequence)
    assert f"invalid {sequence_type} sequence provided" in str(exc_info.value)



from unittest import TestCase

from mavedb.lib.validation.keywords import validate_keyword, validate_keywords
from mavedb.lib.validation.exceptions import ValidationError


class TestKeywordValidators(TestCase):
    """
    Tests that each validator throws the appropriate :class:`ValidationError`
    when passed invalid input.
    """

    def test_ve_invalid_keyword(self):
        with self.assertRaises(ValidationError):
            validate_keyword(555)

    def test_ve_invalid_keyword_in_list(self):
        with self.assertRaises(ValidationError):
            validate_keywords(["protein", 555])



import logging
from operator import attrgetter
from typing import Any, Optional

from fastapi import APIRouter, Depends, HTTPException
from fastapi.encoders import jsonable_encoder
from sqlalchemy.orm import Session

from mavedb import deps
from mavedb.lib.authentication import get_current_user
from mavedb.lib.authorization import require_current_user
from mavedb.lib.experiments import search_experiments as _search_experiments
from mavedb.lib.identifiers import (
    find_or_create_doi_identifier,
    find_or_create_pubmed_identifier,
    find_or_create_raw_read_identifier,
)
from mavedb.models.experiment import Experiment
from mavedb.models.scoreset import Scoreset
from mavedb.models.user import User
from mavedb.view_models import experiment, scoreset
from mavedb.view_models.search import ExperimentsSearch

logger = logging.getLogger(__name__)

router = APIRouter(prefix="/api/v1", tags=["experiments"], responses={404: {"description": "Not found"}})


@router.get("/experiments/", status_code=200, response_model=list[experiment.Experiment])
def list_experiments(
    *,
    editable: Optional[bool] = None,
    q: Optional[str] = None,
    db: Session = Depends(deps.get_db),
    user: User = Depends(get_current_user),
) -> list[Experiment]:
    """
    List experiments.
    """
    query = db.query(Experiment)
    if q is not None:
        if user is None:
            logger.error("USER IS NONE")
            return []
        if len(q) > 0:
            logger.error("Here")
            logger.error(user.id)
            query = query.filter(Experiment.created_by_id == user.id)  # .filter(Experiment.published_date is None)
        # else:
        #     query = query.filter(Experiment.created_by_id == user.id).filter(Experiment.published_date is None)
    items = query.order_by(Experiment.urn).all()
    logger.error(len(items))
    return items


@router.post("/experiments/search", status_code=200, response_model=list[experiment.ShortExperiment])
def search_experiments(search: ExperimentsSearch, db: Session = Depends(deps.get_db)) -> Any:
    """
    Search experiments.
    """
    return _search_experiments(db, None, search)


@router.post("/me/experiments/search", status_code=200, response_model=list[experiment.ShortExperiment])
def search_my_experiments(
    search: ExperimentsSearch, db: Session = Depends(deps.get_db), user: User = Depends(require_current_user)
) -> Any:
    """
    Search experiments created by the current user..
    """
    return _search_experiments(db, user, search)


@router.get("/experiments/{urn}", status_code=200, response_model=experiment.Experiment, responses={404: {}})
def fetch_experiment(*, urn: str, db: Session = Depends(deps.get_db)) -> Experiment:
    """
    Fetch a single experiment by URN.
    """
    # item = db.query(Experiment).filter(Experiment.urn == urn).filter(Experiment.private.is_(False)).first()
    item = db.query(Experiment).filter(Experiment.urn == urn).first()
    if not item:
        raise HTTPException(status_code=404, detail=f"Experiment with URN {urn} not found")
    return item


@router.get(
    "/experiments/{urn}/scoresets",
    status_code=200,
    response_model=list[scoreset.Scoreset],
    responses={404: {}},
)
def get_experiment_scoresets(*, urn: str, db: Session = Depends(deps.get_db)) -> Any:
    """
    Get all score sets belonging to an experiment.
    """
    experiment = db.query(Experiment).filter(Experiment.urn == urn).first()
    if not experiment:
        raise HTTPException(status_code=404, detail=f"Experiment with URN {urn} not found")
    # Only get published score sets. Unpublished score sets won't be shown on experiment page.
    scoresets = (
        db.query(Scoreset).filter(Scoreset.experiment_id == experiment.id).filter(Scoreset.urn.contains("urn")).all()
    )
    if not scoresets:
        raise HTTPException(status_code=404, detail="No associated score set")
    else:
        scoresets.sort(key=attrgetter("urn"))
    return scoresets


@router.post("/experiments/", response_model=experiment.Experiment, responses={422: {}})
async def create_experiment(
    *,
    item_create: experiment.ExperimentCreate,
    db: Session = Depends(deps.get_db),
    user: User = Depends(require_current_user),
) -> Any:
    """
    Create an experiment.
    """
    if item_create is None:
        return None
    doi_identifiers = [
        await find_or_create_doi_identifier(db, identifier.identifier)
        for identifier in item_create.doi_identifiers or []
    ]
    pubmed_identifiers = [
        await find_or_create_pubmed_identifier(db, identifier.identifier)
        for identifier in item_create.pubmed_identifiers or []
    ]
    raw_read_identifiers = [
        await find_or_create_raw_read_identifier(db, identifier.identifier)
        for identifier in item_create.raw_read_identifiers or []
    ]
    item = Experiment(
        **jsonable_encoder(
            item_create,
            by_alias=False,
            exclude=["doi_identifiers", "keywords", "pubmed_identifiers", "raw_read_identifiers"],
        ),
        doi_identifiers=doi_identifiers,
        pubmed_identifiers=pubmed_identifiers,
        raw_read_identifiers=raw_read_identifiers,
        created_by=user,
        modified_by=user,
    )
    await item.set_keywords(db, item_create.keywords)
    db.add(item)
    db.commit()
    db.refresh(item)
    return item


@router.put("/experiments/{urn}", response_model=experiment.Experiment, responses={422: {}})
async def update_experiment(
    *,
    item_update: experiment.ExperimentUpdate,
    urn: str,
    db: Session = Depends(deps.get_db),
    user: User = Depends(require_current_user),
) -> Any:
    """
    Update an experiment.
    """
    if item_update is None:
        return None
    # item = db.query(Experiment).filter(Experiment.urn == urn).filter(Experiment.private.is_(False)).one_or_none()
    item = db.query(Experiment).filter(Experiment.urn == urn).one_or_none()
    if item is None:
        return None

    pairs = {
        k: v
        for k, v in vars(item_update).items()
        if k not in ["doi_identifiers", "keywords", "pubmed_identifiers", "raw_read_identifiers"]
    }
    for var, value in pairs.items():  # vars(item_update).items():
        setattr(item, var, value) if value else None

    doi_identifiers = [
        await find_or_create_doi_identifier(db, identifier.identifier)
        for identifier in item_update.doi_identifiers or []
    ]
    pubmed_identifiers = [
        await find_or_create_pubmed_identifier(db, identifier.identifier)
        for identifier in item_update.pubmed_identifiers or []
    ]
    raw_read_identifiers = [
        await find_or_create_raw_read_identifier(db, identifier.identifier)
        for identifier in item_update.raw_read_identifiers or []
    ]
    item.doi_identifiers = doi_identifiers
    item.pubmed_identifiers = pubmed_identifiers
    item.raw_read_identifiers = raw_read_identifiers

    await item.set_keywords(db, item_update.keywords)
    item.modified_by = user

    db.add(item)
    db.commit()
    db.refresh(item)
    return item


@router.delete("/experiments/{urn}", response_model=experiment.Experiment, responses={422: {}})
async def delete_experiment(
    *, urn: str, db: Session = Depends(deps.get_db), user: User = Depends(require_current_user)
) -> Any:
    """
    Delete a experiment .

    Raises

    Returns
    _______
    Does not return anything
    string : HTTP code 200 successful but returning content
    or
    communitcate to client whether the operation succeeded
    204 if successful but not returning content - likely going with this
    """
    item = db.query(Experiment).filter(Experiment.urn == urn).one_or_none()
    if not item:
        raise HTTPException(status_code=404, detail=f"Experiment with URN {urn} not found.")

    db.delete(item)
    db.commit()



from pydantic.types import Optional

from mavedb.view_models.base.base import BaseModel


class ExperimentsSearch(BaseModel):
    published: Optional[bool]
    text: Optional[str]


class ScoresetsSearch(BaseModel):
    published: Optional[bool]
    targets: Optional[list[str]]
    target_organism_names: Optional[list[str]]
    target_types: Optional[list[str]]
    text: Optional[str]


class TextSearch(BaseModel):
    text: Optional[str]



import datetime
from decimal import Decimal


NON_FIELD_ERRORS = "__all__"

_PROTECTED_TYPES = (type(None), int, float, Decimal, datetime.datetime, datetime.date, datetime.time)


def is_protected_type(obj):
    """Determine if the object instance is of a protected type.

    Objects of protected types are preserved as-is when passed to
    force_str(strings_only=True).
    """
    return isinstance(obj, _PROTECTED_TYPES)


def force_str(s, encoding="utf-8", strings_only=False, errors="strict"):
    """
    Return a str object representing arbitrary object s.

    Bytestrings are treated using the encoding codec.

    If strings_only is True, some non-string-like objects are not converted.
    """
    # Handle the common case first for performance reasons.
    if issubclass(type(s), str):
        return s
    if strings_only and is_protected_type(s):
        return s
    try:
        if isinstance(s, bytes):
            s = str(s, encoding, errors)
        else:
            s = str(s)
    except UnicodeDecodeError as e:
        # raise DjangoUnicodeDecodeError(s, *e.args)
        raise e
    return s


class ValidationError(Exception):
    """An error while validating data."""

    def __init__(self, message, code=None, params=None):
        """
        The `message` argument can be a single error, a list of errors, or a
        dictionary that maps field names to lists of errors. What we define as
        an "error" can be either a simple string or an instance of
        ValidationError with its message attribute set, and what we define as
        list or dictionary can be an actual `list` or `dict` or an instance
        of ValidationError with its `error_list` or `error_dict` attribute set.
        """

        # PY2 can't pickle naive exception: http://bugs.python.org/issue1692335.
        super(ValidationError, self).__init__(message, code, params)

        if isinstance(message, ValidationError):
            if hasattr(message, "error_dict"):
                message = message.error_dict
            # PY2 has a `message` property which is always there so we can't
            # duck-type on it. It was introduced in Python 2.5 and already
            # deprecated in Python 2.6.
            elif not hasattr(message, "message" if six.PY3 else "code"):
                message = message.error_list
            else:
                message, code, params = message.message, message.code, message.params

        if isinstance(message, dict):
            self.error_dict = {}
            for field, messages in message.items():
                if not isinstance(messages, ValidationError):
                    messages = ValidationError(messages)
                self.error_dict[field] = messages.error_list

        elif isinstance(message, list):
            self.error_list = []
            for message in message:
                # Normalize plain strings to instances of ValidationError.
                if not isinstance(message, ValidationError):
                    message = ValidationError(message)
                if hasattr(message, "error_dict"):
                    self.error_list.extend(sum(message.error_dict.values(), []))
                else:
                    self.error_list.extend(message.error_list)

        else:
            self.message = message
            self.code = code
            self.params = params
            self.error_list = [self]

    @property
    def message_dict(self):
        # Trigger an AttributeError if this ValidationError
        # doesn't have an error_dict.
        getattr(self, "error_dict")

        return dict(self)

    @property
    def messages(self):
        if hasattr(self, "error_dict"):
            return sum(dict(self).values(), [])
        return list(self)

    def update_error_dict(self, error_dict):
        if hasattr(self, "error_dict"):
            for field, error_list in self.error_dict.items():
                error_dict.setdefault(field, []).extend(error_list)
        else:
            error_dict.setdefault(NON_FIELD_ERRORS, []).extend(self.error_list)
        return error_dict

    def __iter__(self):
        if hasattr(self, "error_dict"):
            for field, errors in self.error_dict.items():
                yield field, list(ValidationError(errors))
        else:
            for error in self.error_list:
                message = error.message
                if error.params:
                    message %= error.params
                yield force_str(message)

    def __str__(self):
        if hasattr(self, "error_dict"):
            return repr(dict(self))
        return repr(list(self))

    def __repr__(self):
        return "ValidationError(%s)" % self


class OpenIDConnectException(Exception):
    """Raised when OpenID login flow fails."""

    pass



from datetime import date

from pydantic.types import Optional

from .base.base import BaseModel


class ReferenceGenomeBase(BaseModel):
    short_name: str
    organism_name: str
    genome_id: Optional[int]
    creation_date: date
    modification_date: date


class ReferenceGenomeCreate(ReferenceGenomeBase):
    pass


class ReferenceGenomeUpdate(ReferenceGenomeBase):
    pass


# Properties shared by models stored in DB
class ReferenceGenomeInDbBase(ReferenceGenomeBase):
    id: int

    class Config:
        orm_mode = True


# Properties to return to client
class ReferenceGenome(ReferenceGenomeInDbBase):
    pass


# Properties stored in DB
class ReferenceGenomeInDb(ReferenceGenomeInDbBase):
    pass



from unittest import TestCase

from mavedb.lib.validation.target import validate_sequence_category, validate_target_category, validate_target_sequence
from mavedb.lib.validation.exceptions import ValidationError
from mavedb.lib.validation.constants.target import valid_categories, valid_sequence_types


class TestValidateTargetCategory(TestCase):
    def test_valid(self):
        for category in valid_categories:
            validate_target_category(category)

    def test_invalid_category(self):
        with self.assertRaises(ValidationError):
            validate_target_category("Protein")

    def test_invalid_case(self):
        with self.assertRaises(ValidationError):
            validate_target_category("protein coding")


class TestValidateSequenceCategory(TestCase):
    def test_valid(self):
        for sequence_type in valid_sequence_types:
            validate_sequence_category(sequence_type)

    def test_invalid_category(self):
        with self.assertRaises(ValidationError):
            validate_sequence_category("RNA")

    def test_invalid_case(self):
        with self.assertRaises(ValidationError):
            validate_sequence_category("DNA")


class TestValidateTargetSequence(TestCase):
    def test_valid(self):
        validate_target_sequence("ATGACCAAACAT", "dna")
        validate_target_sequence("STARTREK", "protein")

    def test_infer(self):
        validate_target_sequence("ATGACCAAACAT", "infer")
        validate_target_sequence("STARTREK", "infer")

    def test_type_mismatch(self):
        with self.assertRaises(ValidationError):
            validate_target_sequence("STARTREK", "dna")

    def test_invalid_characters(self):
        with self.assertRaises(ValidationError):
            validate_target_sequence("AUGACCAAACAU", "dna")
        with self.assertRaises(ValidationError):
            validate_target_sequence("AUGACCAAACAU", "infer")



from datetime import date

from sqlalchemy import Boolean, Column, Date, ForeignKey, Integer
from sqlalchemy.orm import relationship, backref

from mavedb.db.base import Base


class ReferenceMap(Base):
    __tablename__ = "reference_maps"

    id = Column(Integer, primary_key=True, index=True)
    is_primary = Column(Boolean, nullable=False, default=False)
    genome_id = Column(Integer, ForeignKey("reference_genomes.id"), nullable=False)
    genome = relationship("ReferenceGenome", backref="reference_maps")
    target_id = Column(Integer, ForeignKey("target_genes.id"), nullable=False)
    target = relationship("TargetGene", backref=backref("reference_maps", cascade="all,delete-orphan"))
    creation_date = Column(Date, nullable=False, default=date.today)
    modification_date = Column(Date, nullable=False, default=date.today, onupdate=date.today)



from functools import partial
from typing import Optional, Union

from mavedb.lib.exceptions import ValidationError

from mavehgvs import Variant, MaveHgvsParseError

from .utils import is_csv_null
from .constants import HGVS_NT_COLUMN, HGVS_PRO_COLUMN, HGVS_SPLICE_COLUMN


def validate_hgvs_string(
    value: Union[str, bytes],
    column: Optional[str] = None,
    splice_present: bool = False,
    targetseq: Optional[str] = None,
    relaxed_ordering: bool = False,
) -> Optional[str]:
    if is_csv_null(value):
        return None

    if hasattr(value, "decode"):
        value = value.decode()
    if not isinstance(value, str):
        raise ValidationError(
            "Variant HGVS values input must be strings. " "'{}' has the type '{}'.".format(value, type(value).__name__)
        )

    if value.lower() == "_sy":
        raise ValidationError("_sy is no longer supported and should be replaced by p.(=)")
    elif value.lower() == "_wt":
        raise ValidationError("_wt is no longer supported and should be replaced by (cgnp).=")

    try:
        variant = Variant(s=value, targetseq=targetseq, relaxed_ordering=relaxed_ordering)
    except MaveHgvsParseError as error:
        raise ValidationError(f"{value}: {str(error)}")

    prefix = variant.prefix.lower()
    if column in ("nt", HGVS_NT_COLUMN):
        if splice_present:
            if prefix not in "g":
                raise ValidationError(
                    f"'{value}' is not a genomic variant (prefix 'g.'). "
                    f"Nucleotide variants must be genomic if transcript "
                    f"variants are also defined."
                )
        else:
            if prefix not in "cn":
                raise ValidationError(
                    f"'{value}' is not a transcript variant. The accepted "
                    f"transcript variant prefixes are 'c.', 'n.'."
                )
    elif column in ("splice", HGVS_SPLICE_COLUMN):
        if prefix not in "cn":
            raise ValidationError(
                f"'{value}' is not a transcript variant. The accepted " f"transcript variant prefixes are 'c.', 'n.'."
            )
    elif column in ("p", HGVS_PRO_COLUMN):
        if prefix not in "p":
            raise ValidationError(
                f"'{value}' is not a protein variant. The accepted " f"protein variant prefix is 'p.'."
            )
    else:
        raise ValueError("Unknown column '{}'. Expected nt, splice or p".format(column))

    return str(variant)


validate_nt_variant = partial(validate_hgvs_string, **{"column": "nt"})
validate_splice_variant = partial(validate_hgvs_string, **{"column": "splice"})
validate_pro_variant = partial(validate_hgvs_string, **{"column": "p"})



import uuid
from mavedb.lib.validation.exceptions import ValidationError
from mavedb.lib.validation.constants.urn import (
    MAVEDB_ANY_URN_RE,
    MAVEDB_EXPERIMENT_URN_RE,
    MAVEDB_EXPERIMENTSET_URN_RE,
    MAVEDB_SCORESET_URN_RE,
    MAVEDB_VARIANT_URN_RE,
)


def validate_mavedb_urn(urn: str):
    """
    This function validates a MaveDB urn and raises an error if it is not valid.

    Parameters
    __________
    urn : str
        The MaveDB urn to be validated.

    Raises
    ______
    ValidationError
        If the MaveDB urn is not valid.
    """
    if urn.startswith("tmp:"):
        try:
            uuid.UUID(urn[4:])
        except ValueError:
            raise ValidationError("{}'s is not a valid urn.".format(urn))
    else:
        if not MAVEDB_ANY_URN_RE.match(urn):
            raise ValidationError("{}'s is not a valid urn.".format(urn))


def validate_mavedb_urn_experimentset(urn: str):
    """
    This function validates a Experiment Set urn and raises an error if it is not valid.

    Parameters
    __________
    urn : str
        The Experiment Set urn to be validated.

    Raises
    ______
    ValidationError
        If the Experiment Set urn is not valid.
    """
    if urn.startswith("tmp:"):
        try:
            uuid.UUID(urn[4:])
        except ValueError:
            raise ValidationError("{}'s is not a valid Experiment Set urn.".format(urn))
    else:
        if not MAVEDB_EXPERIMENTSET_URN_RE.match(urn):
            raise ValidationError("{}'s is not a valid Experiment Set urn.".format(urn))


def validate_mavedb_urn_experiment(urn: str):
    """
    This function validates an Experiment urn and raises an error if it is not valid.

    Parameters
    __________
    urn : str
        The Experiment urn to be validated.

    Raises
    ______
    ValidationError
        If the Experiemnt urn is not valid.
    """
    if urn.startswith("tmp:"):
        try:
            uuid.UUID(urn[4:])
        except ValueError:
            raise ValidationError("{}'s is not a valid Experiment Set urn.".format(urn))
    else:
        if not MAVEDB_EXPERIMENT_URN_RE.match(urn):
            raise ValidationError("{}'s is not a valid Experiment urn.".format(urn))


def validate_mavedb_urn_scoreset(urn: str):
    """
    This function validates a Scoreset urn and raises an error if it is not valid.

    Parameters
    __________
    urn : str
        The Scoreset urn to be validated

    Raises
    ______
    ValidationError
        If the Scoreset urn is not valid.
    """
    if urn.startswith("tmp:"):
        try:
            uuid.UUID(urn[4:])
        except ValueError:
            raise ValidationError("{}'s is not a valid score set urn.".format(urn))
    else:
        if not MAVEDB_SCORESET_URN_RE.match(urn):
            raise ValidationError("{}'s is not a valid score set urn.".format(urn))


def validate_mavedb_urn_variant(urn: str):
    """
    This function validates a MaveDB Variant urn and raises an error if it is not valid.

    Parameters
    __________
    urn : str
        The MaveDB Variant urn to be validated.

    Raises
    ______
    ValidationError
        If the MaveDB Variant urn is not valid.
    """
    if urn.startswith("tmp:"):
        try:
            uuid.UUID(urn[4:])
        except ValueError:
            raise ValidationError("{}'s is not a valid Experiment Set urn.".format(urn))
    else:
        if not MAVEDB_VARIANT_URN_RE.match(urn):
            raise ValidationError("{}'s is not a valid Variant urn.".format(urn))



import pandas as pd
from mavehgvs.variant import Variant
from mavehgvs.exceptions import MaveHgvsParseError
import numpy as np
from typing import Optional, Tuple

from mavedb.lib.validation.constants.general import (
    hgvs_nt_column,
    hgvs_splice_column,
    hgvs_pro_column,
    required_score_column,
)
from mavedb.lib.validation.exceptions import ValidationError
from fqfa.util.translate import translate_dna

# handle with pandas all null strings
# provide a csv or a pandas dataframe
# take dataframe, output as csv to temp directory, use standard library


STANDARD_COLUMNS = (hgvs_nt_column, hgvs_splice_column, hgvs_pro_column, required_score_column)


def infer_column_type(col: pd.Series) -> str:
    """Infer whether the given column contains string or numeric data.

    The function returns "string" for string columns or "numeric" for numeric columns.
    If there is a mixture of types it returns "mixed".
    If every value in the column is `None` or NA it returns "empty".

    Parameters
    ----------
    col : pandas.Series
        The column to inspect

    Returns
    -------
    str
        One of "string", "numeric", "mixed", or "empty"
    """
    if col.isna().all():
        return "empty"
    else:
        col_numeric = pd.to_numeric(col, errors="coerce")
        if col_numeric.isna().all():  # nothing converted to a number
            return "string"
        elif np.all(col.isna() == col_numeric.isna()):  # all non-NA values converted
            return "numeric"
        else:  # some values converted but not all
            return "mixed"


def sort_dataframe_columns(df: pd.DataFrame) -> pd.DataFrame:
    """Sort the columns of the given dataframe according to the expected ordering in MaveDB.

    MaveDB expects that dataframes have columns in the following order (note some columns are optional):
    * hgvs_nt
    * hgvs_splice
    * hgvs_pro
    * score
    * other

    Parameters
    ----------
    df : pandas.DataFrame
        The dataframe with columns to sort

    Returns
    -------
    pandas.DataFrame
        The dataframe with the same data but sorted columns
    """

    def column_sort_function(value, columns):
        if value.lower() in STANDARD_COLUMNS:
            return STANDARD_COLUMNS.index(value.lower())
        else:
            return columns.index(value) + len(STANDARD_COLUMNS)

    old_columns = list(df.columns)
    new_columns = sorted(old_columns, key=lambda v: column_sort_function(v, old_columns))

    return df[new_columns]


def standardize_dataframe(df: pd.DataFrame) -> pd.DataFrame:
    """Standardize a dataframe by sorting the columns and changing the standard column names to lowercase.

    The standard column names are:
    * hgvs_nt
    * hgvs_splice
    * hgvs_pro
    * score

    Case for other columns is preserved.

    Parameters
    ----------
    df : pandas.DataFrame
        The dataframe to standardize

    Returns
    -------
    pandas.DataFrame
        The standardized dataframe
    """
    new_columns = [x.lower() if x.lower() in STANDARD_COLUMNS else x for x in df.columns]
    df.columns = new_columns

    return sort_dataframe_columns(df)


def validate_dataframe(df: pd.DataFrame, kind: str, target_seq: str, target_seq_type: str) -> None:
    """
    Validate that a given dataframe passes all checks.

    Parameters
    ----------
    df : pandas.DataFrame
        The dataframe to validate
    kind : str
        The kind of dataframe "counts" or "scores"
    target_seq : str
        The target sequence to validate variants against
    target_seq_type : str
        The kind of target sequence, one of "infer" "dna" or "protein"

    Returns
    -------
    None

    Raises
    ------
    ValidationError
        If one of the validators called raises an exception
    """
    # basic checks
    validate_column_names(df, kind)
    validate_no_null_rows(df)

    # validate individual columns
    column_mapping = {c.lower(): c for c in df.columns}
    index_column = choose_dataframe_index_column(df)
    for c in column_mapping:
        if c in (hgvs_nt_column, hgvs_splice_column, hgvs_pro_column):
            is_index = column_mapping[c] == index_column
            if is_index or not df[column_mapping[c]].isna().all():  # ignore null non-index HGVS columns
                validate_hgvs_column(df[column_mapping[c]], is_index, target_seq, target_seq_type)
        else:
            force_numeric = (c == required_score_column) or (kind == "counts")
            validate_data_column(df[column_mapping[c]], force_numeric)

    # validate hgvs agreement
    prefixes = dict()
    for c in (hgvs_nt_column, hgvs_splice_column, hgvs_pro_column):
        prefixes[c] = None
        if c in column_mapping:
            if not df[column_mapping[c]].isna().all():
                prefixes[c] = df[column_mapping[c]].dropna()[0][0]
    validate_hgvs_prefix_combinations(
        hgvs_nt=prefixes[hgvs_nt_column], hgvs_splice=prefixes[hgvs_splice_column], hgvs_pro=prefixes[hgvs_pro_column]
    )


def validate_and_standardize_dataframe_pair(
    scores_df: pd.DataFrame, counts_df: Optional[pd.DataFrame], target_seq: str, target_seq_type: str
) -> Tuple[pd.DataFrame, Optional[pd.DataFrame]]:
    """
    Perform validation and standardization on a pair of score and count dataframes.

    Parameters
    ----------
    scores_df : pandas.DataFrame
        The scores dataframe
    counts_df : Optional[pandas.DataFrame]
        The counts dataframe, can be None if not present
    target_seq : str
        The target sequence for the dataset
    target_seq_type : str
        The target sequence type, can be "infer" "dna" or "protein"

    Returns
    -------
    Tuple[pd.DataFrame, Optional[pd.DataFrame]]
        The standardized score and count dataframes, or score and None if no count dataframe was provided

    Raises
    ------
    ValidationError
        If one of the validation functions raises an exception
    """
    # validate the dataframes
    validate_dataframe(scores_df, "scores", target_seq, target_seq_type)
    if counts_df is not None:
        validate_dataframe(counts_df, "counts", target_seq, target_seq_type)
        validate_variant_columns_match(scores_df, counts_df)

    new_scores_df = standardize_dataframe(scores_df)
    if counts_df is not None:
        new_counts_df = standardize_dataframe(counts_df)
    else:
        new_counts_df = None
    return new_scores_df, new_counts_df


def choose_dataframe_index_column(df: pd.DataFrame) -> str:
    """
    Identify the HGVS variant column that should be used as the index column in this dataframe.

    Parameters
    ----------
    df : pandas.DataFrame
        The dataframe to check

    Returns
    -------
    str
        The column name of the index column

    Raises
    ------
    ValidationError
        If no valid HGVS variant column is found
    """
    column_mapping = {c.lower(): c for c in df.columns if not df[c].isna().all()}

    if hgvs_nt_column in column_mapping:
        return column_mapping[hgvs_nt_column]
    elif hgvs_pro_column in column_mapping:
        return column_mapping[hgvs_pro_column]
    else:
        raise ValidationError("failed to find valid HGVS variant column")


def validate_no_null_rows(df: pd.DataFrame) -> None:
    """Check that there are no fully null rows in the dataframe.

    Parameters
    __________
    df : pandas.DataFrame
        The scores or counts dataframe being validated

    Raises
    ______
    ValidationError
        If there are null rows in the dataframe
    """
    null_rows = df.apply(lambda row: np.all(row.isna()), axis=1)
    if sum(null_rows) > 0:
        raise ValidationError(f"found {sum(null_rows)} null rows in the data frame")


def validate_column_names(df: pd.DataFrame, kind: str) -> None:
    """Validate the column names in a dataframe.

    This function validates the column names in the given dataframe.
    It can be run for either a "scores" dataframe or a "counts" dataframe.
    A "scores" dataframe must have a column named 'score' and a "counts" dataframe cannot have a column named 'score'.

    The function also checks for a valid combination of columns that define variants.

    Basic checks are performed to make sure that a column name is not empty, null, or whitespace,
    as well as making sure there are no duplicate column names.

    Parameters
    ----------
    df : pandas.DataFrame
        The scores or counts dataframe to be validated

    kind : str
        Either "counts" or "scores" depending on the kind of dataframe being validated

    Raises
    ------
    ValidationError
        If the column names are not valid
    """
    if any(type(c) != str for c in df.columns):
        raise ValidationError("column names must be strings")

    if any(c.isspace() for c in df.columns) or any(len(c) == 0 for c in df.columns):
        raise ValidationError("column names cannot be empty or whitespace")

    columns = [c.lower() for c in df.columns]

    if kind == "scores":
        if required_score_column not in columns:
            raise ValidationError(f"score dataframe must have a '{required_score_column}' column")
    elif kind == "counts":
        if required_score_column in columns:
            raise ValidationError(f"counts dataframe must not have a '{required_score_column}' column")
    else:
        raise ValueError("kind only accepts scores and counts")

    if hgvs_splice_column in columns:
        if hgvs_nt_column not in columns or hgvs_pro_column not in columns:
            raise ValidationError(
                f"dataframes with '{hgvs_splice_column}' must also define '{hgvs_nt_column}' and '{hgvs_pro_column}'"
            )

    if len(columns) != len(set(columns)):
        raise ValidationError("duplicate column names are not allowed (this check is case insensitive)")

    if set(columns).isdisjoint({hgvs_nt_column, hgvs_splice_column, hgvs_pro_column}):
        raise ValidationError("dataframe does not define any variant columns")

    if set(columns).issubset({hgvs_nt_column, hgvs_splice_column, hgvs_pro_column}):
        raise ValidationError("dataframe does not define any data columns")


def validate_hgvs_column(column: pd.Series, is_index: bool, target_seq: str, target_seq_type) -> None:
    """
    Validate the variants in an HGVS column from a dataframe.

    Tests whether the column has a correct and consistent prefix.
    This function also validates all individual variants in the column and checks for agreement against the target
    sequence (for non-splice variants).

    Parameters
    ----------
    column : pd.Series
        The column from the dataframe to validate
    is_index : bool
        True if this is the index column for the dataframe and therefore cannot have missing values; else False
    target_seq : str
        The target sequence to validate against
    target_seq_type : str
        The type of target sequence, which can be one of "dna", "protein", or "infer"

    Returns
    -------
    None

    Raises
    ------
    ValueError
        If the target sequence does is not dna or protein (or inferred as dna or protein)
    ValueError
        If the target sequence is not valid for the variants (e.g. protein sequence for nucleotide variants)
    ValueError
        If the column name is not recognized
    ValidationError
        If the column contains multiple prefixes or the wrong prefix for that column name
    ValidationError
        If an index column contains missing values
    ValidationError
        If one of the variants fails validation
    """
    if target_seq_type not in ("dna", "protein"):
        raise ValueError("invalid target sequence type")

    if infer_column_type(column) not in ("string", "empty"):
        raise ValidationError(f"variant column '{column.name}' cannot contain numeric data")

    if is_index:
        if column.isna().any():
            raise ValidationError(f"primary variant column '{column.name}' cannot contain null values")
        if not column.is_unique:
            raise ValidationError(f"primary variant column '{column.name}' must contain unique values")

    # check the variant prefixes
    if column.name.lower() == hgvs_nt_column:
        prefixes = [f"{a}." for a in "cngmo"]
    elif column.name.lower() == hgvs_splice_column:
        prefixes = [f"{a}." for a in "cn"]
    elif column.name.lower() == hgvs_pro_column:
        prefixes = ["p."]
    else:
        raise ValueError(f"unrecognized hgvs column name '{column.name}'")
    if len(set(s[:2] for s in column.dropna())) > 1:
        raise ValidationError(f"variant column '{column.name}' has inconsistent variant prefixes")
    if not all(s[:2] in prefixes for s in column.dropna()):
        raise ValidationError(f"variant column '{column.name}' has invalid variant prefixes")

    # validate the individual variant strings
    # prepare the target sequence for validation
    if column.name.lower() == hgvs_nt_column:
        if target_seq_type != "dna":
            raise ValueError(f"invalid target sequence type for '{column.name}'")
    elif column.name.lower() == hgvs_splice_column:
        target_seq = None  # don't validate splice variants against the target sequence
    elif column.name.lower() == hgvs_pro_column:
        if target_seq_type == "dna":
            target_seq = translate_dna(target_seq)[0]  # translate the target sequence if needed
    else:
        raise ValueError(f"unrecognized hgvs column name '{column.name}'")
    # get a list of all invalid variants
    invalid_variants = list()
    for i, s in column.items():
        if s is not None:
            try:
                Variant(s, targetseq=target_seq)
            except MaveHgvsParseError:
                try:
                    Variant(s)  # note this will get called a second time for splice variants
                except MaveHgvsParseError:
                    invalid_variants.append(f"invalid variant string '{s}' at row {i}")
                else:
                    invalid_variants.append(f"target sequence mismatch for '{s}' at row {i}")
    # format and raise an error message that contains all invalid variants
    if len(invalid_variants) > 0:
        raise ValidationError(
            f"encountered {len(invalid_variants)} invalid variant strings: {(', '.join(invalid_variants))}"
        )


def validate_hgvs_prefix_combinations(
    hgvs_nt: Optional[str], hgvs_splice: Optional[str], hgvs_pro: Optional[str]
) -> None:
    """
    Validate the combination of HGVS variant prefixes.

    This function assumes that other validation, such as checking that all variants in the column have the same prefix,
    has already been performed.

    Parameters
    ----------
    hgvs_nt : Optional[str]
        The first character (prefix) of the HGVS nucleotide variant strings, or None if not used.
    hgvs_splice : Optional[str]
        The first character (prefix) of the HGVS splice variant strings, or None if not used.
    hgvs_pro : Optional[str]
        The first character (prefix) of the HGVS protein variant strings, or None if not used.

    Returns
    -------
    None

    Raises
    ------
    ValueError
        If upstream validation failed and an invalid prefix string was passed to this function
    ValidationError
        If the combination of prefixes is not valid
    """
    # ensure that the prefixes are valid - this validation should have been performed before this function was called
    if hgvs_nt not in list("cngmo") + [None]:
        raise ValueError("invalid nucleotide prefix")
    if hgvs_splice not in list("cn") + [None]:
        raise ValueError("invalid nucleotide prefix")
    if hgvs_pro not in ["p", None]:
        raise ValueError("invalid protein prefix")

    # test agreement of prefixes across columns
    if hgvs_splice is not None:
        if hgvs_nt not in list("gmo"):
            raise ValidationError("nucleotide variants must use valid genomic prefix when splice variants are present")
        if hgvs_pro is not None:
            if hgvs_splice != "c":
                raise ValidationError("splice variants' must use 'c.' prefix when protein variants are present")
        else:
            if hgvs_splice != "n":
                raise ValidationError("splice variants must use 'n.' prefix when protein variants are not present")
    elif hgvs_pro is not None and hgvs_nt is not None:
        if hgvs_nt != "c":
            raise ValidationError(
                "nucleotide variants must use 'c.' prefix when protein variants are present and splicing variants are"
                " not present"
            )
    elif hgvs_nt is not None:  # just hgvs_nt
        if hgvs_nt != "n":
            raise ValidationError("nucleotide variants must use 'n.' prefix when only nucleotide variants are defined")


def validate_variant_consistency(df: pd.DataFrame) -> None:
    """
    Ensure that variants defined in a single row describe the same variant.

    Parameters
    ----------
    df : pd.DataFrame

    Returns
    -------
    None

    """
    # TODO
    pass


def validate_data_column(column: pd.Series, force_numeric: bool = False) -> None:
    """
    Validate the contents of a data column.

    Parameters
    ----------
    column : pandas.Series
        A data column from a dataframe
    force_numeric : bool
        Force the data to be numeric, used for score column and count data

    Returns
    -------
    None

    Raises
    ------
    ValidationError
        If the data is all null
    ValidationError
        If the data is of mixed numeric and string types
    ValidationError
        If the data is not numeric and force_numeric is True

    """
    column_type = infer_column_type(column)
    if column_type == "empty":
        raise ValidationError(f"data column '{column.name}' contains no data")
    elif column_type == "mixed":
        raise ValidationError(f"data column '{column.name}' has mixed string and numeric types")
    elif force_numeric and column_type != "numeric":
        raise ValidationError(f"data column '{column.name}' must contain only numeric data")


def validate_variant_columns_match(df1: pd.DataFrame, df2: pd.DataFrame):
    """
    Checks if two dataframes have matching HGVS columns.

    The check performed is order-independent.
    This function is used to validate a pair of scores and counts dataframes that were uploaded together.

    Parameters
    ----------
    df1 : pandas.DataFrame
        Dataframe parsed from an uploaded scores file
    df2 : pandas.DataFrame
        Dataframe parsed from an uploaded counts file

    Raises
    ------
    ValidationError
        If both dataframes do not define the same variant columns
    ValidationError
        If both dataframes do not define the same variants within each column
    """
    for c in df1.columns:
        if c.lower() in (hgvs_nt_column, hgvs_splice_column, hgvs_pro_column):
            if c not in df2:
                raise ValidationError("both score and count dataframes must define matching HGVS columns")
            elif df1[c].isnull().all() and df2[c].isnull().all():
                continue
            elif np.any(df1[c].sort_values().values != df2[c].sort_values().values):
                raise ValidationError(
                    f"both score and count dataframes must define matching variants, discrepancy found in '{c}'"
                )
    for c in df2.columns:
        if c.lower() in (hgvs_nt_column, hgvs_splice_column, hgvs_pro_column):
            if c not in df1:
                raise ValidationError("both score and count dataframes must define matching HGVS columns")



from datetime import date
from typing import Any

from pydantic.types import Optional

from .base.base import BaseModel


class VariantBase(BaseModel):
    urn: Optional[str]
    data: Any
    scoreset_id: int
    hgvs_nt: Optional[str]
    hgvs_pro: Optional[str]
    hgvs_splice: Optional[str]
    creation_date: date
    modification_date: date


class VariantCreate(VariantBase):
    pass


class VariantUpdate(VariantBase):
    pass


# Properties shared by models stored in DB
class VariantInDbBase(VariantBase):
    id: int

    class Config:
        orm_mode = True


# Properties to return to client
class Variant(VariantInDbBase):
    pass


# Properties stored in DB
class VariantInDb(VariantInDbBase):
    pass



from datetime import date
from sqlalchemy import Boolean, Column, Date, Enum, ForeignKey, Integer, String
from sqlalchemy.orm import relationship, backref
from sqlalchemy.schema import Table

from mavedb.db.base import Base
from mavedb.deps import JSONB
from mavedb.models.enums.processing_state import ProcessingState
from .keyword import Keyword

# from .raw_read_identifier import SraIdentifier
from mavedb.lib.temp_urns import generate_temp_urn

# TODO Reformat code without removing dependencies whose use is not detected.

scoresets_doi_identifiers_association_table = Table(
    "scoreset_doi_identifiers",
    Base.metadata,
    Column("scoreset_id", ForeignKey("scoresets.id"), primary_key=True),
    Column("doi_identifier_id", ForeignKey("doi_identifiers.id"), primary_key=True),
)


scoresets_keywords_association_table = Table(
    "scoreset_keywords",
    Base.metadata,
    Column("scoreset_id", ForeignKey("scoresets.id"), primary_key=True),
    Column("keyword_id", ForeignKey("keywords.id"), primary_key=True),
)


scoresets_meta_analysis_scoresets_association_table = Table(
    "scoreset_meta_analysis_sources",
    Base.metadata,
    Column("source_scoreset_id", ForeignKey("scoresets.id"), primary_key=True),
    Column("meta_analysis_scoreset_id", ForeignKey("scoresets.id"), primary_key=True),
)


scoresets_pubmed_identifiers_association_table = Table(
    "scoreset_pubmed_identifiers",
    Base.metadata,
    Column("scoreset_id", ForeignKey("scoresets.id"), primary_key=True),
    Column("pubmed_identifier_id", ForeignKey("pubmed_identifiers.id"), primary_key=True),
)


# scoresets_sra_identifiers_association_table = Table(
scoresets_raw_read_identifiers_association_table = Table(
    "scoreset_sra_identifiers",
    Base.metadata,
    Column("scoreset_id", ForeignKey("scoresets.id"), primary_key=True),
    Column("sra_identifier_id", ForeignKey("sra_identifiers.id"), primary_key=True),
)


class Scoreset(Base):
    __tablename__ = "scoresets"

    id = Column(Integer, primary_key=True, index=True)

    urn = Column(String(64), nullable=True, default=generate_temp_urn)  # index=True, nullable=True
    title = Column(String, nullable=False)
    method_text = Column(String, nullable=False)
    abstract_text = Column(String, nullable=False)
    short_description = Column(String, nullable=False)
    extra_metadata = Column(JSONB, nullable=False)
    dataset_columns = Column(JSONB, nullable=False, default={})

    normalised = Column(Boolean, nullable=False, default=False)
    private = Column(Boolean, nullable=False, default=True)
    approved = Column(Boolean, nullable=False, default=False)
    published_date = Column(Date, nullable=True)
    processing_state = Column(
        Enum(ProcessingState, create_constraint=True, length=32, native_enum=False, validate_strings=True),
        nullable=True,
    )
    data_usage_policy = Column(String, nullable=True)

    # TODO Refactor the way we track the number of variants?
    num_variants = Column(Integer, nullable=False, default=0)

    experiment_id = Column(Integer, ForeignKey("experiments.id"), nullable=False)
    experiment = relationship("Experiment", backref=backref("scoresets", cascade="all,delete-orphan"))
    # TODO Standardize on US or GB spelling for licenc/se.
    licence_id = Column(Integer, ForeignKey("licenses.id"), nullable=False)
    license = relationship("License")
    superseded_scoreset_id = Column("replaces_id", Integer, ForeignKey("scoresets.id"), nullable=True)  # TODO
    superseded_scoreset = relationship(
        "Scoreset", uselist=False, remote_side=[id], backref=backref("superseding_scoreset", uselist=False)
    )

    created_by_id = Column(Integer, ForeignKey("users.id"), nullable=True)
    created_by = relationship("User", foreign_keys="Scoreset.created_by_id")
    modified_by_id = Column(Integer, ForeignKey("users.id"), nullable=True)
    modified_by = relationship("User", foreign_keys="Scoreset.modified_by_id")
    creation_date = Column(Date, nullable=False, default=date.today)
    modification_date = Column(Date, nullable=False, default=date.today, onupdate=date.today)

    keyword_objs = relationship("Keyword", secondary=scoresets_keywords_association_table, backref="scoresets")
    doi_identifiers = relationship(
        "DoiIdentifier", secondary=scoresets_doi_identifiers_association_table, backref="scoresets"
    )
    pubmed_identifiers = relationship(
        "PubmedIdentifier", secondary=scoresets_pubmed_identifiers_association_table, backref="scoresets"
    )
    # sra_identifiers = relationship('SraIdentifier', secondary=scoresets_sra_identifiers_association_table, backref='scoresets')
    # raw_read_identifiers = relationship('RawReadIdentifier', secondary=scoresets_raw_read_identifiers_association_table,backref='scoresets')
    meta_analysis_source_scoresets = relationship(
        "Scoreset",
        secondary=scoresets_meta_analysis_scoresets_association_table,
        primaryjoin=(scoresets_meta_analysis_scoresets_association_table.c.meta_analysis_scoreset_id == id),
        secondaryjoin=(scoresets_meta_analysis_scoresets_association_table.c.source_scoreset_id == id),
        backref="meta_analyses",
    )

    # Unfortunately, we can't use association_proxy here, because in spite of what the documentation seems to imply, it
    # doesn't check for a pre-existing keyword with the same text.
    # keywords = association_proxy('keyword_objs', 'text', creator=lambda text: Keyword(text=text))

    # _updated_keywords: list[str] = None
    # _updated_doi_identifiers: list[str] = None

    @property
    def keywords(self) -> list[str]:
        # if self._updated_keywords:
        #     return self._updated_keywords
        # else:
        keyword_objs = self.keyword_objs or []  # getattr(self, 'keyword_objs', [])
        return list(map(lambda keyword_obj: keyword_obj.text, keyword_objs))

    async def set_keywords(self, db, keywords: list[str]):
        self.keyword_objs = [await self._find_or_create_keyword(db, text) for text in keywords]

    # See https://gist.github.com/tachyondecay/e0fe90c074d6b6707d8f1b0b1dcc8e3a
    # @keywords.setter
    # async def set_keywords(self, db, keywords: list[str]):
    #     self._keyword_objs = [await self._find_or_create_keyword(text) for text in keywords]

    async def _find_or_create_keyword(self, db, keyword_text):
        keyword_obj = db.query(Keyword).filter(Keyword.text == keyword_text).one_or_none()
        if not keyword_obj:
            keyword_obj = Keyword(text=keyword_text)
            # object_session.add(keyword_obj)
        return keyword_obj



import secrets
from typing import Any

from cryptography.hazmat.backends import default_backend as crypto_default_backend
from cryptography.hazmat.primitives import serialization as crypto_serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from fastapi import APIRouter, Depends
from fastapi.encoders import jsonable_encoder
from sqlalchemy.orm import Session

from mavedb import deps
from mavedb.lib.authorization import require_current_user
from mavedb.models.access_key import AccessKey
from mavedb.models.user import User
from mavedb.view_models import access_key

router = APIRouter(prefix="/api/v1", tags=["access keys"], responses={404: {"description": "Not found"}})


def generate_key_pair():
    key = rsa.generate_private_key(backend=crypto_default_backend(), public_exponent=65537, key_size=2048)

    private_key = key.private_bytes(
        crypto_serialization.Encoding.PEM, crypto_serialization.PrivateFormat.PKCS8, crypto_serialization.NoEncryption()
    )

    public_key = key.public_key().public_bytes(
        crypto_serialization.Encoding.OpenSSH, crypto_serialization.PublicFormat.OpenSSH
    )

    return private_key, public_key


@router.get(
    "/users/me/access-keys", status_code=200, response_model=list[access_key.AccessKey], responses={404: {}, 500: {}}
)
def list_my_access_keys(*, user: User = Depends(require_current_user)) -> Any:
    """
    List the current user's access keys.
    """
    return user.access_keys


@router.post(
    "/users/me/access-keys", status_code=200, response_model=access_key.NewAccessKey, responses={404: {}, 500: {}}
)
def create_my_access_key(*, db: Session = Depends(deps.get_db), user: User = Depends(require_current_user)) -> Any:
    """
    Create a new access key for the current user.
    """
    private_key, public_key = generate_key_pair()

    item = AccessKey(user=user, key_id=secrets.token_urlsafe(32), public_key=public_key)
    db.add(item)
    db.commit()
    db.refresh(item)

    response_item = access_key.NewAccessKey(**jsonable_encoder(item, by_alias=False), private_key=private_key)
    response_item.private_key = private_key
    return response_item


@router.delete("/users/me/access-keys/{key_id}", status_code=200, responses={404: {}, 500: {}})
def delete_my_access_key(
    *, key_id: str, db: Session = Depends(deps.get_db), user: User = Depends(require_current_user)
) -> Any:
    """
    Delete one of the current user's access keys.
    """
    item = db.query(AccessKey).filter(AccessKey.key_id == key_id).one_or_none()
    if item.user.id == user.id:
        db.delete(item)
    db.commit()



from datetime import datetime

from sqlalchemy import Column, Date, DateTime, Integer, String, ForeignKey
from sqlalchemy.orm import backref, relationship

from mavedb.db.base import Base


class AccessKey(Base):
    __tablename__ = "access_keys"

    id = Column(Integer, primary_key=True)
    user_id = Column(Integer, ForeignKey("users.id"), nullable=False)
    user = relationship("User", backref=backref("access_keys", cascade="all,delete-orphan"))
    key_id = Column(String, unique=True, index=True, nullable=False)
    public_key = Column(String, nullable=False)
    name = Column(String, nullable=True)
    expiration_date = Column(Date, nullable=True)
    creation_time = Column(DateTime, nullable=True, default=datetime.now)



from datetime import date
import json
import re

from tests.conftest import client


def test_create_experiment(test_empty_db):
    experiment_to_create = {
        "title": "Test Experiment Title",
        "methodText": "Methods",
        "abstractText": "Abstract",
        "shortDescription": "Test experiment",
        "extraMetadata": {"key": "value"},
        "keywords": [],
    }
    response = client.post("/api/v1/experiments/", json=experiment_to_create)
    assert response.status_code == 200
    response_data = response.json()
    experiment_urn = response_data["urn"]
    experiment_set_urn = response_data["experimentSetUrn"]
    assert experiment_urn is not None
    assert experiment_set_urn is not None
    assert re.match(r"tmp:[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}", experiment_urn)
    assert re.match(r"tmp:[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}", experiment_set_urn)
    added_fields = {
        "urn": experiment_urn,
        "experimentSetUrn": experiment_set_urn,
        "numScoresets": 0,
        "creationDate": date.today().isoformat(),
        "modificationDate": date.today().isoformat(),
        "publishedDate": None,
        "createdBy": {"orcidId": "someuser", "firstName": "First", "lastName": "Last", "email": None, "roles": []},
        "modifiedBy": {"orcidId": "someuser", "firstName": "First", "lastName": "Last", "email": None, "roles": []},
        "processingState": None,
        "doiIdentifiers": [],
        "pubmedIdentifiers": [],
        "rawReadIdentifiers": [],
    }
    expected_response = experiment_to_create | added_fields
    assert json.dumps(response_data, sort_keys=True) == json.dumps(expected_response, sort_keys=True)


def test_create_experiment_with_invalid_doi(test_empty_db):
    experiment_to_create = {
        "title": "Test Experiment Title",
        "methodText": "Methods",
        "abstractText": "Abstract",
        "shortDescription": "Test experiment",
        "extraMetadata": {"key": "value"},
        "doiIdentifiers": [{"identifier": "10.bf"}],
        "pubmedIdentifiers": [],
        "keywords": [],
    }
    response = client.post("/api/v1/experiments/", json=experiment_to_create)
    assert response.status_code == 422


def test_create_experiment_with_invalid_pubmed(test_empty_db):
    experiment_to_create = {
        "title": "Test Experiment Title",
        "methodText": "Methods",
        "abstractText": "Abstract",
        "shortDescription": "Test experiment",
        "extraMetadata": {"key": "value"},
        "doiIdentifiers": [],
        "pubmedIdentifiers": [{"identifier": "cccc"}],
        "keywords": [],
    }
    response = client.post("/api/v1/experiments/", json=experiment_to_create)
    assert response.status_code == 422



from unittest import TestCase

from mavedb.lib.validation.urn import (
    validate_mavedb_urn,
    validate_mavedb_urn_experiment,
    validate_mavedb_urn_experimentset,
    validate_mavedb_urn_scoreset,
    validate_mavedb_urn_variant,
    ValidationError,
)


class TestValidateUrn(TestCase):
    def test_valid_mavedb_urn(self):
        validate_mavedb_urn("urn:mavedb:00000002-a-1")

    def test_invalid_mavedb_urn(self):
        with self.assertRaises(ValidationError):
            validate_mavedb_urn("urn:mavedb:00000002-a-1-z")

    def test_valid_mavedb_urn_experimentset(self):
        validate_mavedb_urn_experimentset("urn:mavedb:00000001")

    def test_invalid_mavedb_urn_experimentset(self):
        with self.assertRaises(ValidationError):
            validate_mavedb_urn_experimentset("")

    def test_valid_mavedb_urn_experiment(self):
        validate_mavedb_urn_experiment("urn:mavedb:00000001-a")

    def test_invalid_mavedb_urn_experiment(self):
        with self.assertRaises(ValidationError):
            validate_mavedb_urn_experiment("")

    def test_valid_mavedb_urn_scoreset(self):
        validate_mavedb_urn_scoreset("urn:mavedb:00000001-a-1")

    def test_invalid_mavedb_urn_scoreset(self):
        with self.assertRaises(ValidationError):
            validate_mavedb_urn_scoreset("")

    def test_valid_mavedb_urn_variant(self):
        # TODO find a valid variant urn
        pass
        # validate_mavedb_urn_variant("")

    def test_invalid_mavedb_urn_variant(self):
        with self.assertRaises(ValidationError):
            validate_mavedb_urn_variant("urn:mavedb:00000002-a-1")  # this is a scoreset urn


class TestValidateTmpUrn(TestCase):
    def test_valid_tmp_mavedb_urn(self):
        validate_mavedb_urn("tmp:0a56b8eb-8e19-4906-8cc7-d17d884330a5")

    def test_invalid_tmp_mavedb_urn(self):
        with self.assertRaises(ValidationError):
            validate_mavedb_urn("urn:mavedb:00000002-a-1-z")

    def test_valid_tmp_mavedb_urn_experimentset(self):
        validate_mavedb_urn_experimentset("tmp:0a56b8eb-8e19-4906-8cc7-d17d884330a5")

    def test_invalid_tmp_mavedb_urn_experimentset(self):
        with self.assertRaises(ValidationError):
            validate_mavedb_urn_experimentset("")

    def test_valid_tmp_mavedb_urn_experiment(self):
        validate_mavedb_urn_experiment("urn:mavedb:00000001-a")

    def test_invalid_tmp_mavedb_urn_experiment(self):
        with self.assertRaises(ValidationError):
            validate_mavedb_urn_experiment("")

    def test_valid_tmp_mavedb_urn_scoreset(self):
        validate_mavedb_urn_scoreset("tmp:0a56b8eb-8e19-4906-8cc7-d17d884330a5")

    def test_invalid_tmp_mavedb_urn_scoreset(self):
        with self.assertRaises(ValidationError):
            validate_mavedb_urn_scoreset("")

    def test_valid_tmp_mavedb_urn_variant(self):
        validate_mavedb_urn_variant("tmp:0a56b8eb-8e19-4906-8cc7-d17d884330a5")

    def test_invalid_tmp_mavedb_urn_variant(self):
        with self.assertRaises(ValidationError):
            validate_mavedb_urn_variant("urn:mavedb:00000002-a-1")  # this is a scoreset urn



from datetime import date

from sqlalchemy import Column, Date, ForeignKey, Integer, String
from sqlalchemy.orm import relationship, backref

from mavedb.db.base import Base
from mavedb.deps import JSONB


class Variant(Base):
    __tablename__ = "variants"

    id = Column(Integer, primary_key=True, index=True)

    urn = Column(String(64), nullable=True)  # index=True, nullable=True
    data = Column(JSONB, nullable=False)

    scoreset_id = Column(Integer, ForeignKey("scoresets.id"), nullable=False)
    # TODO examine if delete-orphan is necessary, explore cascade
    scoreset = relationship("Scoreset", backref=backref("variants", cascade="all,delete-orphan"))

    hgvs_nt = Column(String, nullable=True)
    hgvs_pro = Column(String, nullable=True)
    hgvs_splice = Column(String, nullable=True)

    creation_date = Column(Date, nullable=False, default=date.today)
    modification_date = Column(Date, nullable=False, default=date.today, onupdate=date.today)



from sqlalchemy import Column, Integer, String

from mavedb.db.base import Base


class Role(Base):
    __tablename__ = "roles"

    id = Column(Integer, primary_key=True)
    name = Column(String, index=True, nullable=False)



from sqlalchemy import Boolean, Column, DateTime, ForeignKey, Integer, String, Table
from sqlalchemy.orm import relationship

from mavedb.db.base import Base
from mavedb.models.role import Role

users_roles_association_table = Table(
    "users_roles",
    Base.metadata,
    Column("user_id", ForeignKey("users.id"), primary_key=True),
    Column("role_id", ForeignKey("roles.id"), primary_key=True),
)


class User(Base):
    __tablename__ = "users"

    id = Column(Integer, primary_key=True)
    username = Column(String, index=True, nullable=False)
    first_name = Column(String, nullable=True)
    last_name = Column(String, nullable=True)
    email = Column(String, nullable=True)
    is_superuser = Column(Boolean, nullable=False, default=False)
    is_staff = Column(Boolean, nullable=False, default=False)
    is_active = Column(Boolean, nullable=False)
    date_joined = Column(DateTime, nullable=True)
    email = Column(String, nullable=True)
    role_objs = relationship("Role", secondary=users_roles_association_table, backref="users")
    last_login = Column(DateTime, nullable=True)

    @property
    def roles(self) -> list[str]:
        role_objs = self.role_objs or []
        return list(map(lambda role_obj: role_obj.name, role_objs))

    async def set_roles(self, db, roles: list[str]):
        self.role_objs = [await self._find_or_create_role(db, name) for name in roles]

    @staticmethod
    async def _find_or_create_role(db, role_name):
        role_obj = db.query(Role).filter(Role.name == role_name).one_or_none()
        if not role_obj:
            role_obj = Role(name=role_name)
            # object_session.add(role_obj)
        return role_obj



from datetime import date
from typing import Any, Dict, Optional

from mavedb.view_models.base.base import BaseModel, validator
from mavedb.view_models.doi_identifier import DoiIdentifierCreate, SavedDoiIdentifier, DoiIdentifier
from mavedb.view_models.pubmed_identifier import PubmedIdentifierCreate, SavedPubmedIdentifier, PubmedIdentifier
from mavedb.view_models.raw_read_identifier import RawReadIdentifierCreate, SavedRawReadIdentifier, RawReadIdentifier
from mavedb.view_models.user import SavedUser, User
from mavedb.lib.validation import keywords


class ExperimentBase(BaseModel):
    title: str
    short_description: Optional[str]
    abstract_text: Optional[str]
    method_text: Optional[str]
    extra_metadata: Dict
    keywords: Optional[list[str]]

    @classmethod
    def from_orm(cls, obj: Any):
        try:
            obj.experiment_set_urn = obj.experiment_set.urn
        except AttributeError:
            obj.experiment_set_urn = None
        return super().from_orm(obj)

    # @validator('urn')
    # def name_must_contain_space(cls, v):
    #    if ' ' not in v:
    #        raise ValueError('must contain a space')
    #    return v.title()


class ExperimentModify(ExperimentBase):
    @validator("keywords")
    def validate_keywords(cls, v):
        keywords.validate_keywords(v)
        return v


class ExperimentCreate(ExperimentModify):
    short_description: str
    abstract_text: str
    method_text: str
    doi_identifiers: Optional[list[DoiIdentifierCreate]]
    pubmed_identifiers: Optional[list[PubmedIdentifierCreate]]
    raw_read_identifiers: Optional[list[RawReadIdentifierCreate]]


class ExperimentUpdate(ExperimentModify):
    short_description: str
    abstract_text: str
    method_text: str
    doi_identifiers: Optional[list[DoiIdentifierCreate]]
    pubmed_identifiers: Optional[list[PubmedIdentifierCreate]]
    raw_read_identifiers: Optional[list[RawReadIdentifierCreate]]


# Properties shared by models stored in DB
class SavedExperiment(ExperimentBase):
    urn: str
    num_scoresets: int
    created_by: Optional[SavedUser]
    modified_by: Optional[SavedUser]
    creation_date: date
    modification_date: date
    published_date: Optional[date]
    experiment_set_urn: Optional[str]
    doi_identifiers: list[SavedDoiIdentifier]
    pubmed_identifiers: list[SavedPubmedIdentifier]
    raw_read_identifiers: list[SavedRawReadIdentifier]
    processing_state: Optional[str]

    class Config:
        orm_mode = True


# Properties to return to non-admin clients
class Experiment(SavedExperiment):
    doi_identifiers: list[DoiIdentifier]
    pubmed_identifiers: list[PubmedIdentifier]
    raw_read_identifiers: list[RawReadIdentifier]
    created_by: Optional[User]
    modified_by: Optional[User]


class ShortExperiment(Experiment):
    pass


# Properties to return to admin clients
class AdminExperiment(Experiment):
    approved: bool



import re

NA_VALUE = "NA"

NULL_VALUES = ("", "na", "nan", "nil", "none", "null", "n/a", "undefined", NA_VALUE)


NULL_VALUES_RE = re.compile("|".join(NULL_VALUES), flags=re.IGNORECASE)
# NULL_VALUES_RE = re.compile(fr'|none|nan|na|undefined|n/a|null|nil|{NA_VALUE}', flags=re.IGNORECASE)


READABLE_NULL_VALUES = [f"'{v}'".format(v) for v in set([v.lower() for v in NULL_VALUES]) if v.strip()] + ["whitespace"]


# html_null_values = [
#     f"<b>{v.strip().lower() or 'whitespace'}</b>" for v in null_values_list
# ]
# HUMANIZED_NULL_VALUES = (
#     f'{", ".join(html_null_values[:-1])} ' f"and " f"{html_null_values[-1]}"
# )


def is_csv_null(value):
    """Return True if a string from a CSV file represents a NULL value."""
    # Number 0 is treated as False so that all 0 will be converted to NA value.
    if value == 0:
        return value
    return not value or NULL_VALUES_RE.fullmatch(str(value).strip().lower())



from humps.camel import case
from pydantic import BaseModel as PydanticBaseModel, validator


class BaseModel(PydanticBaseModel):
    @validator("*", pre=True)
    def empty_str_to_none(cls, x):
        """
        Convert empty strings to None. This is applied to all string-valued attributes before other validators run.

        :param x: The attribute value
        :return: None if x was the empty string, otherwise x
        """
        if x == "":
            return None
        return x

    class Config:
        alias_generator = case
        allow_population_by_field_name = True



from datetime import date

from mavedb.view_models.base.base import BaseModel, validator
from mavedb.lib.validation import target

from fqfa import infer_sequence_type


class WildTypeSequenceBase(BaseModel):
    sequence_type: str
    sequence: str


class WildTypeSequenceModify(WildTypeSequenceBase):
    @validator("sequence_type")
    def validate_category(cls, field_value, values, field, config):
        field_value = field_value.lower()
        target.validate_sequence_category(field_value)
        if field_value == "infer":
            if "sequence" in values.keys():
                field_value = infer_sequence_type(values["sequence"])
            else:
                raise ValueError("sequence is invalid")
        return field_value

    @validator("sequence")
    def validate_identifier(cls, field_value, values, field, config):
        # If sequence_type is invalid, values["sequence_type"] doesn't exist.
        field_value = field_value.upper()
        if "sequence_type" in values.keys():
            sequence_type = values["sequence_type"]
            # field_value is sequence
            target.validate_target_sequence(field_value, sequence_type)
        else:
            raise ValueError("sequence_type is invalid")
        return field_value


class WildTypeSequenceCreate(WildTypeSequenceModify):
    pass


class WildTypeSequenceUpdate(WildTypeSequenceModify):
    pass


# Properties shared by models stored in DB
class SavedWildTypeSequence(WildTypeSequenceBase):
    class Config:
        orm_mode = True
        arbitrary_types_allowed = True


# Properties to return to non-admin clients
class WildTypeSequence(SavedWildTypeSequence):
    pass


# Properties to return to admin clients
class AdminWildTypeSequence(SavedWildTypeSequence):
    creation_date: date
    modification_date: date



"""empty message

Revision ID: 988ca84c701b
Revises: 9d566d915a2c
Create Date: 2023-04-10 05:19:28.099693

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '988ca84c701b'
down_revision = '9d566d915a2c'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('licenses',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('long_name', sa.String(), nullable=False),
    sa.Column('short_name', sa.String(), nullable=False),
    sa.Column('text', sa.String(), nullable=False),
    sa.Column('link', sa.String(), nullable=True),
    sa.Column('version', sa.String(), nullable=True),
    sa.Column('creation_date', sa.Date(), nullable=False),
    sa.Column('modification_date', sa.Date(), nullable=False),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('long_name'),
    sa.UniqueConstraint('short_name')
    )
    op.create_index(op.f('ix_licenses_id'), 'licenses', ['id'], unique=False)
    op.alter_column('scoresets', 'licence_id',
               existing_type=sa.INTEGER(),
               nullable=False)
    op.create_foreign_key(None, 'scoresets', 'licenses', ['licence_id'], ['id'])
    op.create_foreign_key(None, 'scoresets', 'scoresets', ['replaces_id'], ['id'])
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(None, 'scoresets', type_='foreignkey')
    op.drop_constraint(None, 'scoresets', type_='foreignkey')
    op.alter_column('scoresets', 'licence_id',
               existing_type=sa.INTEGER(),
               nullable=True)
    op.drop_index(op.f('ix_licenses_id'), table_name='licenses')
    op.drop_table('licenses')
    # ### end Alembic commands ###



import idutils

from mavedb.lib.validation.exceptions import ValidationError
from mavedb.lib.validation.utilities import is_null
from mavedb.lib.validation.constants.identifier import valid_dbnames


def validate_db_name(db_name: str):
    if db_name.strip() == "" or not db_name:
        raise ValidationError("db_name should not be empty")
    if db_name not in valid_dbnames:
        raise ValidationError(
            f"The `db_name` key within the identifier attribute of the external identifier should "
            f"take one of the following values: {valid_dbnames}."
        )


def validate_identifier(db_name: str, identifier: str):
    if db_name == "UniProt":
        validate_uniprot_identifier(identifier)
    elif db_name == "RefSeq":
        validate_refseq_identifier(identifier)
    elif db_name == "Ensembl":
        validate_ensembl_identifier(identifier)


def validate_sra_identifier(identifier: str):
    """
    Validates whether the identifier is a valid SRA identifier.

    Parameters
    __________
    identifier : str
        The identifier to be validated.

    Raises
    ______
    ValidationError
        If the identifier is not a valid SRA identifier.
    """
    if not (
        idutils.is_sra(identifier)
        or idutils.is_bioproject(identifier)
        or idutils.is_geo(identifier)
        or idutils.is_arrayexpress_array(identifier)
        or idutils.is_arrayexpress_experiment(identifier)
    ):
        raise ValidationError(f"'{identifier} is not a valid SRA, GEO, ArrayExpress or BioProject " "accession.")


def validate_ensembl_identifier(identifier: str):
    """
    Validates whether the identifier is a valid Ensembl identifier.

    Parameters
    __________
    identifier : str
        The identifier to be validated.

    Raises
    ______
    ValidationError
        If the identifier is not a valid Ensembl identifier.
    """
    if not idutils.is_ensembl(identifier):
        raise ValidationError(f"'{identifier}' is not a valid Ensembl accession.")


def validate_uniprot_identifier(identifier: str):
    """
    Validates whether the identifier is a valid UniProt identifier.

    Parameters
    __________
    identifier : str
        The identifier to be validated.

    Raises
    ______
    ValidationError
        If the identifier is not a valid UniProt identifier.
    """
    if not idutils.is_uniprot(identifier):
        raise ValidationError(f"'{identifier}' is not a valid UniProt accession.")


def validate_refseq_identifier(identifier: str):
    """
    Validates whether the identifier is a valid RefSeq identifier.

    Parameters
    __________
    identifier : str
        The identifier to be validated.

    Raises
    ______
    ValidationError
        If the identifier is not a valid RefSeq identifier.
    """
    if not idutils.is_refseq(identifier):
        raise ValidationError(f"'{identifier}' is not a valid RefSeq accession.")


def validate_genome_identifier(identifier: str):
    """
    Validates whether the identifier is a valid genome identifier.

    Parameters
    __________
    identifier : str
        The identifier to be validated.

    Raises
    ______
    ValidationError
        If the identifier is not a valid genome identifier.
    """
    if not idutils.is_genome(identifier):
        raise ValidationError(f"'{identifier}' is not a valid GenBank or RefSeq genome assembly.")


def validate_sra_list(values: list[str]):
    """
    Validates whether each identifier in a list of identifiers (values) is a valid SRA identifier.

    Parameters
    __________
    identifier : list[str]
        The list of identifiers to be validated.

    Raises
    ______
    ValidationError
        If at least one of the identifiers is not a valid SRA identifier.
    """
    for value in values:
        if not is_null(value):
            validate_sra_identifier(value)


def validate_ensembl_list(values: list[str]):
    """
    Validates whether each identifier in a list of identifiers (values) is a valid Ensembl identifier.

    Parameters
    __________
    identifier : list[str]
        The list of identifiers to be validated.

    Raises
    ______
    ValidationError
        If at least one of the identifiers is not a valid Ensemble identifier.
    """
    for value in values:
        if not is_null(value):
            validate_ensembl_identifier(value)


def validate_refseq_list(values: list[str]):
    """
    Validates whether each identifier in a list of identifiers (values) is a valid RefSeq identifier.

    Parameters
    __________
    identifier : list[str]
        The list of identifiers to be validated.

    Raises
    ______
    ValidationError
        If at least one of the identifiers is not a valid RefSeq identifier.
    """
    for value in values:
        if not is_null(value):
            validate_refseq_identifier(value)


def validate_uniprot_list(values: list[str]):
    """
    Validates whether each identifer in a list of identifiers (values) is a valid UniProt identifier.

    Parameters
    __________
    identifier : list[str]
        The list of identifiers to be validated.

    Raises
    ______
    ValidationError
        If at least one of the identifiers is not a valid UniProt identifier.
    """
    for value in values:
        if not is_null(value):
            validate_uniprot_identifier(value)



from datetime import date

from sqlalchemy import Boolean, Column, Date, ForeignKey, Integer, String
from sqlalchemy.event import listens_for
from sqlalchemy.orm import relationship, backref
from sqlalchemy.schema import Table

from mavedb.db.base import Base
from mavedb.deps import JSONB
from mavedb.lib.temp_urns import generate_temp_urn
from mavedb.models.experiment_set import ExperimentSet
from mavedb.models.keyword import Keyword

experiments_doi_identifiers_association_table = Table(
    "experiment_doi_identifiers",
    Base.metadata,
    Column("experiment_id", ForeignKey("experiments.id"), primary_key=True),
    Column("doi_identifier_id", ForeignKey("doi_identifiers.id"), primary_key=True),
)


experiments_keywords_association_table = Table(
    "experiment_keywords",
    Base.metadata,
    Column("experiment_id", ForeignKey("experiments.id"), primary_key=True),
    Column("keyword_id", ForeignKey("keywords.id"), primary_key=True),
)


experiments_pubmed_identifiers_association_table = Table(
    "experiment_pubmed_identifiers",
    Base.metadata,
    Column("experiment_id", ForeignKey("experiments.id"), primary_key=True),
    Column("pubmed_identifier_id", ForeignKey("pubmed_identifiers.id"), primary_key=True),
)


# experiments_sra_identifiers_association_table = Table(
experiments_raw_read_identifiers_association_table = Table(
    "experiment_sra_identifiers",
    Base.metadata,
    Column("experiment_id", ForeignKey("experiments.id"), primary_key=True),
    Column("sra_identifier_id", ForeignKey("sra_identifiers.id"), primary_key=True),
)


class Experiment(Base):
    __tablename__ = "experiments"

    id = Column(Integer, primary_key=True, index=True)

    urn = Column(String(64), nullable=True, default=generate_temp_urn)  # index=True, nullable=True
    title = Column(String, nullable=False)
    short_description = Column(String, nullable=False)
    abstract_text = Column(String, nullable=False)
    method_text = Column(String, nullable=False)
    extra_metadata = Column(JSONB, nullable=False)

    private = Column(Boolean, nullable=False, default=True)
    approved = Column(Boolean, nullable=False, default=False)
    published_date = Column(Date, nullable=True)
    processing_state = Column(String, nullable=True)

    # TODO Refactor the way we track the number of scoresets?
    num_scoresets = Column(Integer, nullable=False, default=0)

    experiment_set_id = Column(Integer, ForeignKey("experiment_sets.id"), nullable=True)
    experiment_set = relationship("ExperimentSet", backref=backref("experiments", cascade="all,delete-orphan"))

    created_by_id = Column(Integer, ForeignKey("users.id"), nullable=True)
    created_by = relationship("User", foreign_keys="Experiment.created_by_id")
    modified_by_id = Column(Integer, ForeignKey("users.id"), nullable=True)
    modified_by = relationship("User", foreign_keys="Experiment.modified_by_id")
    creation_date = Column(Date, nullable=False, default=date.today)
    modification_date = Column(Date, nullable=False, default=date.today, onupdate=date.today)

    keyword_objs = relationship("Keyword", secondary=experiments_keywords_association_table, backref="experiments")
    doi_identifiers = relationship(
        "DoiIdentifier", secondary=experiments_doi_identifiers_association_table, backref="experiments"
    )
    pubmed_identifiers = relationship(
        "PubmedIdentifier", secondary=experiments_pubmed_identifiers_association_table, backref="experiments"
    )
    # sra_identifiers = relationship('SraIdentifier', secondary=experiments_sra_identifiers_association_table, backref='experiments')
    raw_read_identifiers = relationship(
        "RawReadIdentifier", secondary=experiments_raw_read_identifiers_association_table, backref="experiments"
    )

    # Unfortunately, we can't use association_proxy here, because in spite of what the documentation seems to imply, it
    # doesn't check for a pre-existing keyword with the same text.
    # keywords = association_proxy('keyword_objs', 'text', creator=lambda text: Keyword(text=text))

    # _updated_keywords: list[str] = None
    # _updated_doi_identifiers: list[str] = None

    @property
    def keywords(self) -> list[str]:
        # if self._updated_keywords:
        #     return self._updated_keywords
        # else:
        keyword_objs = self.keyword_objs or []  # getattr(self, 'keyword_objs', [])
        return list(map(lambda keyword_obj: keyword_obj.text, keyword_objs))

    async def set_keywords(self, db, keywords: list[str]):
        self.keyword_objs = [await self._find_or_create_keyword(db, text) for text in keywords]

    # See https://gist.github.com/tachyondecay/e0fe90c074d6b6707d8f1b0b1dcc8e3a
    # @keywords.setter
    # async def set_keywords(self, db, keywords: list[str]):
    #     self._keyword_objs = [await self._find_or_create_keyword(text) for text in keywords]

    async def _find_or_create_keyword(self, db, keyword_text):
        keyword_obj = db.query(Keyword).filter(Keyword.text == keyword_text).one_or_none()
        if not keyword_obj:
            keyword_obj = Keyword(text=keyword_text)
            # object_session.add(keyword_obj)
        return keyword_obj


@listens_for(Experiment, "before_insert")
def create_parent_object(mapper, connect, target):
    if target.experiment_set_id is None:
        target.experiment_set = ExperimentSet(
            title=target.title,
            method_text=target.method_text,
            abstract_text=target.abstract_text,
            short_description=target.short_description,
            extra_metadata={},
            num_experiments=1,
            created_by=target.created_by,
            modified_by=target.modified_by,
        )



import logging
from datetime import datetime
from typing import Optional

from fastapi import Depends, HTTPException, Request, Security
from fastapi.security import APIKeyCookie, APIKeyHeader, APIKeyQuery, HTTPAuthorizationCredentials, HTTPBearer
from jose import jwt
from sqlalchemy.orm import Session

from mavedb import deps
from mavedb.models.access_key import AccessKey
from mavedb.models.user import User

# See https://8gwifi.org/jwkconvertfunctions.jsp
ORCID_JWT_SIGNING_PUBLIC_KEY = """-----BEGIN PUBLIC KEY-----
MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAjxTIntA7YvdfnYkLSN4w
k//E2zf/wbb0SV/HLHFvh6a9ENVRD1/rHK0EijlBzikb+1rgDQihJETcgBLsMoZV
QqGj8fDUUuxnVHsuGav/bf41PA7E/58HXKPrB2C0cON41f7K3o9TStKpVJOSXBrR
WURmNQ64qnSSryn1nCxMzXpaw7VUo409ohybbvN6ngxVy4QR2NCC7Fr0QVdtapxD
7zdlwx6lEwGemuqs/oG5oDtrRuRgeOHmRps2R6gG5oc+JqVMrVRv6F9h4ja3UgxC
DBQjOVT1BFPWmMHnHCsVYLqbbXkZUfvP2sO1dJiYd/zrQhi+FtNth9qrLLv3gkgt
wQIDAQAB
-----END PUBLIC KEY-----
"""
ORCID_JWT_AUDIENCE = "APP-GXFVWWJT8H0F50WD"

ACCESS_TOKEN_NAME = "X-API-key"

logger = logging.getLogger(__name__)


####################################################################################################
# JWT authentication
####################################################################################################


def decode_jwt(token: str) -> dict:
    try:
        decoded_token = jwt.decode(
            token,
            ORCID_JWT_SIGNING_PUBLIC_KEY,
            algorithms=["RS256"],
            audience=ORCID_JWT_AUDIENCE,
            # ORCID sends an at_hash when using the OpenID Connect implicit flow, even though there is no auth_token.
            options={"verify_at_hash": False},
        )
        return decoded_token
    except Exception as ex:
        print(ex)
        return {}


class JWTBearer(HTTPBearer):
    def __init__(self, auto_error: bool = True):
        super(JWTBearer, self).__init__(auto_error=auto_error)

    async def __call__(self, request: Request):
        credentials: HTTPAuthorizationCredentials
        try:
            credentials = await super(JWTBearer, self).__call__(request)
        except HTTPException:
            credentials = None
        if credentials:
            if not credentials.scheme == "Bearer":
                raise HTTPException(status_code=403, detail="Invalid authentication scheme.")
            token_payload = self.verify_jwt(credentials.credentials)
            if not token_payload:
                raise HTTPException(status_code=403, detail="Invalid token or expired token.")
            return token_payload
        else:
            return None

    @staticmethod
    def verify_jwt(token: str) -> bool:
        try:
            payload = decode_jwt(token)
        except:
            payload = None
        return payload


####################################################################################################
# API key authentication
####################################################################################################

access_token_query = APIKeyQuery(name=ACCESS_TOKEN_NAME, auto_error=False)
access_token_header = APIKeyHeader(name=ACCESS_TOKEN_NAME, auto_error=False)
access_token_cookie = APIKeyCookie(name=ACCESS_TOKEN_NAME, auto_error=False)


async def get_access_token(
        # access_token_query: str = Security(access_token_query),
        access_token_header: Optional[str] = Security(access_token_header),
        access_token_cookie: Optional[str] = Security(access_token_cookie),
) -> Optional[str]:
    return access_token_header or access_token_cookie


async def get_current_user_from_api_key(
        db: Session = Depends(deps.get_db), access_token: str = Depends(get_access_token)
) -> Optional[User]:
    user = None
    if access_token is not None:
        access_key = db.query(AccessKey).filter(AccessKey.key_id == access_token).one_or_none()
        if access_key:
            user = access_key.user
    return user


####################################################################################################
# Main authentication methods
####################################################################################################


async def get_current_user(
        api_key_user: Optional[User] = Depends(get_current_user_from_api_key),
        token_payload: dict = Depends(JWTBearer()),
        db: Session = Depends(deps.get_db),
) -> Optional[User]:
    user = api_key_user
    if user is None and token_payload is not None:
        username: str = token_payload["sub"]
        if username is not None:
            user = db.query(User).filter(User.username == username).one_or_none()
            if user is None:
                # A new user has just connected an ORCID iD. Create the user account.
                user = User(
                    username=username,
                    is_active=True,
                    # TODO When we decouple from the old database, change first_name and last_name to be nullable, and
                    # stop filling them with empty strings.
                    first_name=token_payload["given_name"] if "given_name" in token_payload else "",
                    last_name=token_payload["family_name"] if "family_name" in token_payload else "",
                    date_joined=datetime.now(),
                )
                logger.info(f"Creating new user with username {user.username}")
                db.add(user)
                db.commit()
                db.refresh(user)
            elif not user.is_active:
                user = None
    return user



from celery.utils.log import get_task_logger

from mavedb.deps import get_db
from mavedb.lib.scoresets import create_variants_data
from mavedb.lib.worker import celery_app
from mavedb.models.scoreset import Scoreset
from mavedb.models.variant import Variant

logger = get_task_logger(__name__)


# TODO Currently unused, but kept here for future reference when we need background processing.
# In moving from Django to FastAPI, we eliminated the need to use this for short tasks like scoreset ingestion, thanks
# to FastAPI's support for asynchronous operations. But if any operations are really long-running or computationally
# intensive, we should move them to a separate process and have the option of moving them to a separate server.


@celery_app.task(
    bind=True,
    ignore_result=True,
    # base=BaseCreateVariantsTask,
    serializer="pickle",
)
def create_variants_task(self, scoreset_urn, scores, counts, index_col, dataset_columns, user_id=None):
    """
    Celery task to that creates and associates `variant.model.Variant` instances
    parsed/validated during upload to a `models.scoreset.ScoreSet` instance.

    Parameters
    ----------
    self : `BaseCreateVariantsTask`
        Bound when celery calls this task.
    user_id : int
        Primary key (id) of the submitting user.
    scoreset_urn : str
        The urn of the instance to associate variants to.
    scores : str
        JSON formatted dataframe (NaN replaced with None).
    counts :
        JSON formatted dataframe (NaN replaced with None).
    index_col : str
        HGVS column to use as the index when matching up variant data between
        scores and counts.
    dataset_columns : dict
        Contains keys `scores` and `counts`. The values are lists of strings
        indicating the columns to be expected in the variants for this dataset.

    Returns
    -------
    `models.scoreset.ScoreSet`
    """
    self.urn = scoreset_urn
    # Look for instances. This might throw an ObjectDoesNotExist exception.
    # Bind ORM objects if they were found.
    self.user = None  # User.objects.get(pk=user_id)
    db = next(get_db())
    self.instance = db.query(Scoreset).get(Scoreset.urn == scoreset_urn).one_or_none()

    logger.info(f"Sending scores dataframe with {len(scores)} rows.")
    logger.info(f"Sending counts dataframe with {len(counts)} rows.")
    logger.info(f"Formatting variants for {self.urn}")
    variants = create_variants_data(scores, counts, index_col)

    if variants:
        logger.info(f"{self.urn}:{variants[-1]}")

    # with transaction.atomic():
    logger.info(f"Deleting existing variants for {self.urn}")
    self.instance.delete_variants()
    logger.info(f"Creating variants for {self.urn}")
    Variant.bulk_create(self.instance, variants)
    logger.info(f"Saving {self.urn}")
    self.instance.dataset_columns = dataset_columns
    self.instance.save()

    return self.instance



from datetime import date
from typing import Optional

from mavedb.view_models.base.base import BaseModel


class AccessKeyBase(BaseModel):
    key_id: str
    name: Optional[str]
    expiration_date: Optional[date]
    created_at: Optional[str]


# Properties shared by models stored in DB
class SavedAccessKey(AccessKeyBase):
    class Config:
        orm_mode = True


# Properties to return to non-admin clients
class AccessKey(SavedAccessKey):
    pass


# Properties to return when an access key has just been created
class NewAccessKey(SavedAccessKey):
    private_key: str



from sqlalchemy import Boolean, Column, DateTime, Integer, String

from mavedb.db.base import Base


class AuthCode(Base):
    __tablename__ = "social_auth_code"

    id = Column(Integer, primary_key=True, index=True)
    email = Column(String(254), nullable=False)
    code = Column(String(32), nullable=False)
    verified = Column(Boolean, nullable=False)
    date_created = Column(DateTime, nullable=False)



from datetime import date
from typing import Dict, List

from pydantic.types import Optional

from mavedb.view_models.base.base import BaseModel
from mavedb.view_models.experiment import Experiment, SavedExperiment
from mavedb.view_models.user import SavedUser, User


class ExperimentSetBase(BaseModel):
    urn: str
    title: Optional[str]
    method_text: Optional[str]
    abstract_text: Optional[str]
    short_description: Optional[str]
    extra_metadata: Optional[Dict]
    published_date: Optional[date]


class ExperimentSetCreate(ExperimentSetBase):
    pass


class ExperimentSetUpdate(ExperimentSetBase):
    pass


# Properties shared by models stored in DB
class SavedExperimentSet(ExperimentSetBase):
    id: int
    experiments: List[SavedExperiment]
    num_experiments: int
    created_by: Optional[SavedUser]
    modified_by: Optional[SavedUser]
    creation_date: date
    modification_date: date

    class Config:
        orm_mode = True


# Properties to return to non-admin clients
class ExperimentSet(SavedExperimentSet):
    created_by: Optional[User]
    modified_by: Optional[User]
    experiments: List[Experiment]


# Properties to return to admin clients
class AdminExperimentSet(SavedExperimentSet):
    private: bool
    approved: bool
    processing_state: Optional[str]
    created_by: Optional[User]
    modified_by: Optional[User]
    experiments: List[Experiment]



from mavedb.view_models.external_gene_identifier import ExternalGeneIdentifierCreate
from mavedb.view_models.external_gene_identifier_offset import ExternalGeneIdentifierOffsetCreate

import pytest


def test_create_ensemble_identifier(test_empty_db):
    # Test valid identifier
    db_name = "Ensembl"
    identifier = "ENSG00000103275"
    externalIdentifier = ExternalGeneIdentifierCreate(db_name=db_name, identifier=identifier)
    assert externalIdentifier.db_name == "Ensembl"
    assert externalIdentifier.identifier == "ENSG00000103275"


def test_create_invalid_ensemble_identifier(test_empty_db):
    # Test valid identifier
    db_name = "Ensembl"
    invalid_identifier = "not_an_identifier"
    with pytest.raises(ValueError) as exc_info:
        ExternalGeneIdentifierCreate(db_name=db_name, identifier=invalid_identifier)
    assert "'not_an_identifier' is not a valid Ensembl accession." in str(exc_info.value)


def test_create_uniprot_identifier(test_empty_db):
    db_name = "UniProt"
    identifier = "P63279"
    externalIdentifier = ExternalGeneIdentifierCreate(db_name=db_name, identifier=identifier)
    assert externalIdentifier.db_name == "UniProt"
    assert externalIdentifier.identifier == "P63279"


def test_create_invalid_uniprot_identifier(test_empty_db):
    db_name = "UniProt"
    invalid_identifier = "not_an_identifier"
    with pytest.raises(ValueError) as exc_info:
        ExternalGeneIdentifierCreate(db_name=db_name, identifier=invalid_identifier)
    assert "'not_an_identifier' is not a valid UniProt accession." in str(exc_info.value)


def test_create_refseq_identifier(test_empty_db):
    db_name = "RefSeq"
    identifier = "NM_003345"
    externalIdentifier = ExternalGeneIdentifierCreate(db_name=db_name, identifier=identifier)
    assert externalIdentifier.db_name == "RefSeq"
    assert externalIdentifier.identifier == "NM_003345"


def test_create_invalid_refseq_identifier(test_empty_db):
    db_name = "RefSeq"
    invalid_identifier = "not_an_identifier"
    with pytest.raises(ValueError) as exc_info:
        ExternalGeneIdentifierCreate(db_name=db_name, identifier=invalid_identifier)
    assert "'not_an_identifier' is not a valid RefSeq accession." in str(exc_info.value)


def test_empty_db_name(test_empty_db):
    db_name = ""
    identifier = "ENSG00000103275"
    with pytest.raises(ValueError) as exc_info:
        ExternalGeneIdentifierCreate(db_name=db_name, identifier=identifier)
    assert "none is not an allowed value" in str(exc_info.value)


def test_space_db_name(test_empty_db):
    db_name = "   "
    identifier = "ENSG00000103275"
    with pytest.raises(ValueError) as exc_info:
        ExternalGeneIdentifierCreate(db_name=db_name, identifier=identifier)
    assert "db_name should not be empty" in str(exc_info.value)


def test_none_db_name(test_empty_db):
    db_name = None
    identifier = "ENSG00000103275"
    with pytest.raises(ValueError) as exc_info:
        ExternalGeneIdentifierCreate(db_name=db_name, identifier=identifier)
    assert "none is not an allowed value" in str(exc_info.value)


def test_invalid_db_name(test_empty_db):
    db_name = "Invalid"
    identifier = "ENSG00000103275"
    with pytest.raises(ValueError) as exc_info:
        ExternalGeneIdentifierCreate(db_name=db_name, identifier=identifier)
    assert (
        "The `db_name` key within the identifier attribute of the external identifier should take one of "
        "the following values: ['UniProt', 'RefSeq', 'Ensembl']." in str(exc_info.value)
    )


def test_create_identifier_with_offset(test_empty_db):
    identifier = {"db_name": "RefSeq", "identifier": "NM_003345"}
    externalIdentifier = ExternalGeneIdentifierOffsetCreate(identifier=identifier, offset=1)
    assert externalIdentifier.offset == 1


def test_create_identifier_with_string_offset(test_empty_db):
    identifier = {"db_name": "RefSeq", "identifier": "NM_003345"}
    offset = "invalid"
    with pytest.raises(ValueError) as exc_info:
        ExternalGeneIdentifierOffsetCreate(identifier=identifier, offset=offset)
    assert "value is not a valid integer" in str(exc_info.value)


def test_create_identifier_with_negative_offset(test_empty_db):
    identifier = {"db_name": "RefSeq", "identifier": "NM_003345"}
    with pytest.raises(ValueError) as exc_info:
        ExternalGeneIdentifierOffsetCreate(identifier=identifier, offset=-10)
    assert "Offset should not be a negative number" in str(exc_info.value)



from unittest import TestCase

from mavedb.lib.validation.variant import validate_hgvs_string
from mavedb.lib.validation.exceptions import ValidationError
from mavedb.lib.validation.constants.general import null_values_list


class TestValidateHgvsString(TestCase):
    def test_passes_on_null(self):
        for v in null_values_list:
            validate_hgvs_string(v)

    def test_error_not_str(self):
        with self.assertRaises(ValidationError):
            validate_hgvs_string(1.0)

    def test_error_unknown_column(self):
        with self.assertRaises(ValueError):
            validate_hgvs_string("c.1A>G", column="random")

    def test_error_does_not_match_splice(self):
        with self.assertRaises(ValidationError):
            validate_hgvs_string("g.G4L", column="splice")

    def test_error_nt_is_not_g_when_splice_present(self):
        validate_hgvs_string("c.1A>G", column="nt", splice_present=False)
        with self.assertRaises(ValidationError):
            validate_hgvs_string("c.1A>G", column="nt", splice_present=True)

    def test_error_does_not_match_nt(self):
        with self.assertRaises(ValidationError):
            validate_hgvs_string("p.G4L", column="nt")

    def test_error_does_not_match_pro(self):
        with self.assertRaises(ValidationError):
            validate_hgvs_string("c.1A>G", column="p")

    def test_raises_on_enrich_special_types(self):
        with self.assertRaises(ValidationError):
            validate_hgvs_string("_wt")
        with self.assertRaises(ValidationError):
            validate_hgvs_string("_sy")

    def test_validates_valid_hgvs(self):
        validate_hgvs_string("c.1A>G", column="nt", splice_present=False)
        validate_hgvs_string("g.1A>G", column="nt", splice_present=True)
        validate_hgvs_string("c.1A>G", column="splice")
        validate_hgvs_string("p.(=)", column="p")


class TestHGVSValidator(TestCase):
    """
    Tests the function :func:`validate_hgvs_string` to see if it is able
    to validate strings which do not comply with the HGVS standard for
    coding, non-coding and nucleotide variants and multi-variants.
    """

    def test_validation_error_not_str_or_bytes(self):
        with self.assertRaises(ValidationError):
            validate_hgvs_string([])

    def test_does_not_pass_enrich_wt_hgvs(self):
        with self.assertRaises(ValidationError):
            validate_hgvs_string("_wt")

    def test_does_not_pass_enrich_sy_hgvs(self):
        with self.assertRaises(ValidationError):
            validate_hgvs_string("_sy")

    def test_passes_multi(self):
        validate_hgvs_string("p.[Lys4Gly;Lys5Phe]", column="p")
        validate_hgvs_string("c.[1A>G;127_128delinsAGC]", column="nt")
        validate_hgvs_string("c.[1A>G;127_128delinsAGC]", column="splice")

    def test_error_invalid_hgvs(self):
        with self.assertRaises(ValidationError):
            validate_hgvs_string("c.ad", column="nt")

    def test_error_invalid_nt_prefix(self):
        with self.assertRaises(ValidationError):
            validate_hgvs_string("r.1a>g", column="nt")

        with self.assertRaises(ValidationError):
            validate_hgvs_string("c.1A>G", column="nt", splice_present=True)

    def test_error_invalid_splice_prefix(self):
        with self.assertRaises(ValidationError):
            validate_hgvs_string("r.1a>g", column="splice")

    def test_error_invalid_pro_prefix(self):
        with self.assertRaises(ValidationError):
            validate_hgvs_string("r.1a>g", column="p")

    def test_converts_bytes_to_string_before_validation(self):
        validate_hgvs_string(b"c.427A>G", column="splice")

    def test_return_none_for_null(self):
        for c in null_values_list:
            self.assertIsNone(validate_hgvs_string(c, column="nt"))



from sqlalchemy import Column, Integer, String

from mavedb.db.base import Base


class AuthAssociation(Base):
    __tablename__ = "social_auth_association"

    id = Column(Integer, primary_key=True, index=True)
    server_url = Column(String(255), nullable=False)
    handle = Column(String(255), nullable=False)
    secret = Column(String(255), nullable=False)
    issued = Column(Integer, nullable=False)
    lifetime = Column(Integer, nullable=False)
    assoc_type = Column(String(64), nullable=True)



import re

MAVEDB_EXPERIMENTSET_URN_DIGITS = 8
# MAVEDB_TMP_URN_DIGITS = 16
MAVEDB_URN_MAX_LENGTH = 64
MAVEDB_URN_NAMESPACE = "mavedb"


# Temp URN patterns
# --------------------------------------------------------------------------- #
# TODO get tmp pattern from UUID4 regex
# MAVEDB_TMP_URN_PATTERN = r"^tmp:[A-Za-z0-9]{{{width}}}$".format(
#    width=MAVEDB_TMP_URN_DIGITS
# )
# MAVEDB_TMP_URN_RE = re.compile(MAVEDB_TMP_URN_PATTERN)


# Experimentset Pattern/Compiled RE
MAVEDB_EXPERIMENTSET_URN_PATTERN = r"^urn:{namespace}:\d{{{width}}}$".format(
    namespace=MAVEDB_URN_NAMESPACE, width=MAVEDB_EXPERIMENTSET_URN_DIGITS
)
MAVEDB_EXPERIMENTSET_URN_RE = re.compile(MAVEDB_EXPERIMENTSET_URN_PATTERN)

# Experiment Pattern/Compiled RE
MAVEDB_EXPERIMENT_URN_PATTERN = r"{pattern}-([a-z]+|0)$".format(pattern=MAVEDB_EXPERIMENTSET_URN_PATTERN[:-1])
MAVEDB_EXPERIMENT_URN_RE = re.compile(MAVEDB_EXPERIMENT_URN_PATTERN)

# Scoreset Pattern/Compiled RE
MAVEDB_SCORESET_URN_PATTERN = r"{pattern}-\d+$".format(pattern=MAVEDB_EXPERIMENT_URN_PATTERN[:-1])
MAVEDB_SCORESET_URN_RE = re.compile(MAVEDB_SCORESET_URN_PATTERN)

# Variant Pattern/Compiled RE
MAVEDB_VARIANT_URN_PATTERN = r"{pattern}#\d+$".format(pattern=MAVEDB_SCORESET_URN_PATTERN[:-1])
MAVEDB_VARIANT_URN_RE = re.compile(MAVEDB_VARIANT_URN_PATTERN)

# Any Pattern/Compiled RE
MAVEDB_ANY_URN_PATTERN = "|".join(
    [
        r"({pattern})".format(pattern=p)
        for p in (
            MAVEDB_EXPERIMENTSET_URN_PATTERN,
            MAVEDB_EXPERIMENT_URN_PATTERN,
            MAVEDB_SCORESET_URN_PATTERN,
            MAVEDB_VARIANT_URN_PATTERN,
            # MAVEDB_TMP_URN_PATTERN,
        )
    ]
)
MAVEDB_ANY_URN_RE = re.compile(MAVEDB_ANY_URN_PATTERN)



import re
import string

from sqlalchemy import func
from sqlalchemy.orm import Session

from mavedb.models.experiment import Experiment
from mavedb.models.experiment_set import ExperimentSet
from mavedb.models.scoreset import Scoreset


def generate_experiment_set_urn(db: Session):
    """
    Generate a new URN for an experiment set.

    Experiment set URNs include an 8-digit, zero-padded, sequentially-assigned numeric part. This function finds the
    maximum value in the database and adds one to form the new URN. To ensure atomicity, it should be called in the
    context of a database transaction.

    :param db: An active database session
    :return: The next available experiment set URN
    """

    # PostgreSQL-specific operator: ~
    # TODO Provide an alternative for use with SQLite in unit tests.
    # TODO We can't use func.max if an experiment set URN's numeric part will ever have anything other than 8 digits,
    # because we rely on the order guaranteed by zero-padding. This assumption is valid until we have 99999999
    # experiment sets.
    row = db.query(func.max(ExperimentSet.urn)).filter(ExperimentSet.urn.op("~")("^urn:mavedb:[0-9]+$")).one_or_none()
    max_urn_number = 0
    if row and row[0]:
        max_urn = row[0]
        max_urn_number = int(re.search("^urn:mavedb:([0-9]+)$", max_urn).groups(1)[0])
    next_urn_number = max_urn_number + 1
    return f"urn:mavedb:{next_urn_number:08}"


def generate_experiment_urn(db: Session, experiment_set: ExperimentSet, experiment_is_meta_analysis: bool):
    """
    Generate a new URN for an experiment.

    Experiment URNs include a two sequentially-assigned parts: a numeric part from the parent experiment set and a
    lowercase alphabetic part identifying the experiment within its set. The alphabetic part is assigned as follows:
    ```
    a, b, ..., z, aa, ab, ..., az, ba, ... bz, ... zz, aaa, ..., zzz, aaaa, ...
    ```
    This function looks at the database records for other experiments in the set and finds the maximum value of the
    alphabetic part, then increments it to form the new URN. To ensure atomicity, it should be called in the context of
    a database transaction to ensure atomicity.

    For meta-analyses, the suffix is always 0. There can only be one meta-analysis per experiment set.

    :param db: An active database session
    :param experiment_set: The experiment set to which this experiment belongs
    :param experiment_is_meta_analysis: Whether the experiment is a meta-analysis
    :return: The next available experiment URN
    """

    experiment_set_urn = experiment_set.urn

    if experiment_is_meta_analysis:
        # Do not increment for meta-analysis, since this is a singleton
        next_suffix = "0"
    else:
        # PostgreSQL-specific operator: ~
        # TODO Provide an alternative for use with SQLite in unit tests.
        published_experiments_query = (
            db.query(Experiment)
            .filter(Experiment.experiment_set_id == experiment_set.id)
            .filter(Experiment.urn.op("~")(f"^{re.escape(experiment_set_urn)}-[a-z]+$"))
        )
        max_suffix = None
        for experiment in published_experiments_query:
            suffix = re.search(f"^{re.escape(experiment_set.urn)}-([a-z]+)$", experiment.urn).group(1)
            if suffix and (max_suffix is None or len(max_suffix) < len(experiment.urn) or max_suffix < suffix):
                max_suffix = suffix
        if max_suffix is None:
            next_suffix = "a"
        else:
            max_suffix_number = 0
            while len(max_suffix) > 0:
                max_suffix_number *= 26
                max_suffix_number += string.ascii_lowercase.index(max_suffix[0]) + 1
                max_suffix = max_suffix[1:]
            next_suffix_number = max_suffix_number + 1
            next_suffix = ""
            x = next_suffix_number
            while x > 0:
                x, y = divmod(x - 1, len(string.ascii_lowercase))
                next_suffix = f"{string.ascii_lowercase[y]}{next_suffix}"
    return f"{experiment_set_urn}-{next_suffix}"


def generate_scoreset_urn(db: Session, experiment: Experiment):
    """
    Generate a new URN for a score set.

    Score set URNs append a sequentially-assigned numeric part to their parent experiment URNs. This numeric part is not
    zero-padded to a fixed width.

    This function looks at the database records for other scoresets belonging to the experiment and finds the maximum
    value of the numeric part, then increments it to form the new URN. To ensure atomicity, it should be called in the
    context of a database transaction.

    :param db: An active database session
    :param experiment: The experiment to which this score set belongs
    :return: The next available score set URN
    """

    experiment_urn = experiment.urn

    # PostgreSQL-specific operator: ~
    # TODO Provide an alternative for use with SQLite in unit tests.
    published_scoresets_query = (
        db.query(Scoreset)
        .filter(Scoreset.experiment_id == experiment.id)
        .filter(Scoreset.urn.op("~")(f"^{re.escape(experiment_urn)}-[0-9]+$"))
    )
    max_suffix_number = 0
    for scoreset in published_scoresets_query:
        suffix_number = int(re.search(f"^{re.escape(experiment.urn)}-([0-9]+)$", scoreset.urn).group(1))
        if suffix_number > max_suffix_number:
            max_suffix_number = suffix_number
    next_suffix_number = max_suffix_number + 1
    return f"{experiment_urn}-{next_suffix_number}"
